import os
import posix
import copy

import numpy as np

from ....trees.morphtree import MorphLoc
from ....trees.phystree import PhysTree, PhysNode

import neuron
from neuron import h
h.load_file("stdlib.hoc") # contains the lambda rule
h.nrn_load_dll(os.path.join(os.path.dirname(__file__),
                            'x86_64/.libs/libnrnmech.so')) # load all mechanisms
# neuron.load_mechanisms(os.path.join(os.path.dirname(__file__),
#                             'x86_64/.libs/libnrnmech.so')) # load all mechanisms
# mechanism_name_translation NEURON

class MechName(object):
    def __init__(self):
        self.names = {'L': 'pas', 'ca': 'CaDyn'}

    def __getitem__(self, key):
        if key in self.names:
            return self.names[key]
        else:
            return 'I' + key
            # 'TestChannel': 'ITestChannel', 'TestChannel2': 'ITestChannel2',
            # 'h': 'Ih', 'h_HAY': 'Ih_HAY',
            # 'Na': 'INa', 'Na_p': 'INa_p', 'Na_Ta': 'INa_Ta', 'Na_Ta2': 'INa_Ta2',
            # 'K': 'IK', 'Klva': 'IKlva', 'KA': 'IKA', 'm': 'Im', 'Kpst': 'IKpst', 'Ktst': 'IKtst', 'Kv3_1': 'IKv3_1', 'SK': 'ISK',
            # 'Ca_LVA': 'ICa_LVA', 'Ca_HVA': 'ICa_HVA',
            # 'NaP': 'INaP', 'NaF': 'INaF',
            # 'CaE': 'ICaE', 'CaP': 'ICaP', 'CaP2': 'ICaP2', 'CaT': 'ICaT',
            # 'K23': 'IK23', 'KC3': 'IKC3', 'KA': 'IKA', 'KD': 'IKD', 'KM': 'IKM', 'Kh': 'IKh', 'Khh': 'IKhh',
mechname = MechName()


class NeuronSimNode(PhysNode):
    def __init__(self, index, p3d=None):
        super().__init__(index, p3d)

    def _makeSection(self, factorlambda=1., pprint=False):
        compartment = neuron.h.Section(name=str(self.index))
        compartment.push()
        # create the compartment
        if self.index == 1:
            compartment.diam = 2. * self.R # um (NEURON takes diam=2*r)
            compartment.L = 2. * self.R    # um (to get correct surface)
            compartment.nseg = 1
        else:
            compartment.diam = 2. * self.R  # section radius [um] (NEURON takes diam = 2*r)
            compartment.L = self.L # section length [um]
            # set number of segments
            if type(factorlambda) == float:
                # nseg according to NEURON bookint
                compartment.nseg = int(((compartment.L / (0.1 * h.lambda_f(100.)) + 0.9) / 2.) * 2. + 1.) * int(factorlambda)
            else:
                compartment.nseg = factorlambda

        # set parameters
        compartment.cm = self.c_m # uF/cm^2
        compartment.Ra = self.r_a*1e6 # MOhm*cm --> Ohm*cm
        # insert membrane currents
        for key, current in self.currents.items():
            if current[0] > 1e-10:
                compartment.insert(mechname[key])
                for seg in compartment:
                    exec('seg.' + mechname[key] + '.g = ' + str(current[0]) + '*1e-6') # uS/cm^2 --> S/cm^2
                    exec('seg.' + mechname[key] + '.e = ' + str(current[1])) # mV
        # insert concentration mechanisms
        for ion, params in self.concmechs.items():
            compartment.insert(mechname[ion])
            for seg in compartment:
                for param, value in params.items():
                    exec('seg.' + mechname[ion] + '.' + param + ' = ' + str(value))
        h.pop_section()

        if pprint:
            print(self)
            print(('>>> compartment length = %.2f um'%compartment.L))
            print(('>>> compartment diam = %.2f um'%compartment.diam))
            print(('>>> compartment nseg = ' + str(compartment.nseg)))

        return compartment

    def _makeShunt(self, compartment):
        if self.g_shunt > 1e-10:
            shunt = h.Shunt(compartment(1.))
            shunt.g = self.g_shunt # uS
            shunt.e = self.e_eq    # mV
            return shunt
        else:
            return None


class NeuronSimTree(PhysTree):
    def __init__(self, file_n=None, types=[1,3,4],
                       factor_lambda=1., t_calibrate=0., dt=0.025, v_init=-75.):
        super().__init__(file_n=file_n, types=types)
        # neuron storage
        self.sections = {}
        self.shunts = []
        self.syns = []
        self.iclamps = []
        self.vclamps = []
        self.vecstims = []
        self.netcons = []
        self.vecs = []
        # simulation parameters
        self.dt = dt # ms
        self.t_calibrate = t_calibrate # ms
        self.factor_lambda = factor_lambda
        self.v_init = v_init # mV
        self.indstart = int(t_calibrate / dt)

    def _createCorrespondingNode(self, node_index, p3d=None):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
            node_index: int
                index of the new node
        """
        return NeuronSimNode(node_index, p3d=p3d)

    def initModel(self, dt=0.025, t_calibrate=0., v_init=-75., factor_lambda=1.,
                        pprint=False):
        """
        Initialize hoc-objects to simulate the neuron model implemented by this
        tree.

        Parameters
        ----------
        dt: float (default is ``.025`` ms)
            Timestep of the simulation
        t_calibrate: float (default ``0.`` ms)
            The calibration time; time model runs without input to reach its
            equilibrium state before the true simulation starts
        v_init: float (default ``-75.`` mV)
            The initial voltage at which the model is initialized
        factor_lambda: float or int (default 1.)
            If int, the number of segments per section. If float, multiplies the
            number of segments given by the standard lambda rule (Carnevale, 2004)
            to give the number of compartments simulated (default value 1. gives
            the number given by the lambda rule)
        pprint: bool (default ``False``)
            Whether or not to print info on the NEURON model's creation
        """
        self.t_calibrate = t_calibrate
        self.dt = dt
        self.indstart = int(self.t_calibrate / self.dt)
        self.v_init = v_init
        self.factor_lambda = factor_lambda
        # reset all storage
        self.deleteModel()
        # create the NEURON model
        self._createNeuronTree(pprint=pprint)

    def deleteModel(self):
        '''
        Delete all stored hoc-objects
        '''
        # reset all storage
        self.sections = {}
        self.shunts = []
        self.syns = []
        self.iclamps = []
        self.vclamps = []
        self.vecstims = []
        self.netcons = []
        self.vecs = []
        self.storeLocs([{'node': 1, 'x': 0.}], 'rec locs')

    def _createNeuronTree(self, pprint):
        for node in self:
            # create the NEURON section
            compartment = node._makeSection(self.factor_lambda, pprint=pprint)
            # connect with parent section
            if not self.isRoot(node):
                compartment.connect(self.sections[node.parent_node.index], 1, 0)
            # store
            self.sections.update({node.index: compartment})
            # create a static shunt
            shunt = node._makeShunt(compartment)
            if shunt is not None:
                self.shunts.append(shunt)
        # if pprint:
        #     print(h.topology())


    def addShunt(self, loc, g, e_r):
        """
        Adds a static conductance at a given location

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the shunt.
        g: float
            The conductance of the shunt (uS)
        e_r: float
            The reversal potential of the shunt (mV)
        """
        loc = MorphLoc(loc, self)
        # create the shunt
        shunt = h.Shunt(self.sections[loc['node']](loc['x']))
        shunt.g = g
        shunt.e = e_r
        # store the shunt
        self.shunts.append(shunt)

    def addDoubleExpCurrent(self, loc, tau1, tau2):
        """
        Adds a double exponential input current at a given location

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        tau1: float
            Rise time of the current waveform (ms)
        tau2: float
            Decay time of the current waveform (ms)
        """
        loc = MorphLoc(loc, self)
        # create the synapse
        syn = h.epsc_double_exp(self.sections[loc['node']](loc['x']))
        syn.tau1 = tau1
        syn.tau2 = tau2
        # store the synapse
        self.syns.append(syn)

    def addExpSyn(self, loc, tau, e_r):
        """
        Adds a single-exponential conductance-based synapse

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        tau: float
            Decay time of the conductance window (ms)
        e_r: float
           Reversal potential of the synapse (mV)
        """
        loc = MorphLoc(loc, self)
        # create the synapse
        syn = h.exp_AMPA_NMDA(self.sections[loc['node']](loc['x']))
        syn.tau = tau
        syn.e = e_r
        # store the synapse
        self.syns.append(syn)

    def addDoubleExpSynapse(self, loc, tau1, tau2, e_r):
        """
        Adds a double-exponential conductance-based synapse

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        tau1: float
            Rise time of the conductance window (ms)
        tau2: float
            Decay time of the conductance window (ms)
        e_r: float
            Reversal potential of the synapse (mV)
        """
        loc = MorphLoc(loc, self)
        # create the synapse
        syn = h.Exp2Syn(self.sections[loc['node']](loc['x']))
        syn.tau1 = tau1
        syn.tau2 = tau2
        syn.e = e_r
        # store the synapse
        self.syns.append(syn)

    def addNMDASynapse(self, loc, tau, tau_nmda, e_r=0., nmda_ratio=1.7):
        """
        Adds a single-exponential conductance-based synapse with an AMPA and an
        NMDA component

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        tau: float
            Decay time of the AMPA conductance window (ms)
        tau_nmda: float
            Decay time of the NMDA conductance window (ms)
        e_r: float (optional, default ``0.`` mV)
           Reversal potential of the synapse (mV)
        nmda_ratio: float (optional, default 1.7)
            The ratio of the NMDA over AMPA component. Means that the maximum of
            the NMDA conductance window is ``nmda_ratio`` times the maximum of
            the AMPA conductance window.
        """
        loc = MorphLoc(loc, self)
        # create the synapse
        syn = h.exp_AMPA_NMDA(self.sections[loc['node']](loc['x']))
        syn.tau = tau
        syn.tau_NMDA = tau_nmda
        syn.e = e_r
        syn.NMDA_ratio = nmda_ratio
        # store the synapse
        self.syns.append(syn)

    def addDoubleExpNMDASynapse(self, loc, tau1, tau2, tau1_nmda, tau2_nmda,
                                     e_r=0., nmda_ratio=1.7):
        """
        Adds a double-exponential conductance-based synapse with an AMPA and an
        NMDA component

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        tau1: float
            Rise time of the AMPA conductance window (ms)
        tau2: float
            Decay time of the AMPA conductance window (ms)
        tau1_nmda: float
            Rise time of the NMDA conductance window (ms)
        tau2_nmda: float
            Decay time of the NMDA conductance window (ms)
        e_r: float (optional, default ``0.`` mV)
           Reversal potential of the synapse (mV)
        nmda_ratio: float (optional, default 1.7)
            The ratio of the NMDA over AMPA component. Means that the maximum of
            the NMDA conductance window is ``nmda_ratio`` times the maximum of
            the AMPA conductance window.
        """
        loc = MorphLoc(loc, self)
        # create the synapse
        syn = h.double_exp_AMPA_NMDA(self.sections[loc['node']](loc['x']))
        syn.tau1 = tau1
        syn.tau2 = tau2
        syn.tau1_NMDA = tau1_nmda
        syn.tau2_NMDA = tau2_nmda
        syn.e = e_r
        syn.NMDA_ratio = nmda_ratio
        # store the synapse
        self.syns.append(syn)

    def addIClamp(self, loc, amp, delay, dur):
        """
        Injects a DC current step at a given lcoation

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        amp: float
            The amplitude of the current (nA)
        delay: float
            The delay of the current step onset (ms)
        dur: float
            The duration of the current step (ms)
        """
        loc = MorphLoc(loc, self)
        # create the current clamp
        iclamp = h.IClamp(self.sections[loc['node']](loc['x']))
        iclamp.delay = delay + self.t_calibrate # ms
        iclamp.dur = dur # ms
        iclamp.amp = amp # nA
        # store the iclamp
        self.iclamps.append(iclamp)

    def addSinClamp(self, loc, amp, delay, dur, bias, freq, phase):
        """
        Injects a sinusoidal current at a given lcoation

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        amp: float
            The amplitude of the current (nA)
        delay: float
            The delay of the current onset (ms)
        dur: float
            The duration of the current (ms)
        bias: float
            Constant baseline added to the sinusoidal waveform (nA)
        freq: float
            Frequency of the sinusoid (Hz)
        phase: float
            Phase of the sinusoid (rad)
        """
        loc = MorphLoc(loc, self)
        # create the current clamp
        iclamp = h.SinClamp(self.sections[loc['node']](loc['x']))
        iclamp.delay = delay + self.t_calibrate # ms
        iclamp.dur = dur # ms
        iclamp.pkamp = amp # nA
        iclamp.bias = bias # nA
        iclamp.freq = freq # Hz
        iclamp.phase = phase # rad
        # store the iclamp
        self.iclamps.append(iclamp)

    def addOUClamp(self, loc, tau, mean, stdev, delay, dur, seed=None):
        """
        Injects a Ornstein-Uhlenbeck current at a given lcoation

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the current.
        tau: float
            Time-scale of the OU process (ms)
        mean: float
            Mean of the OU process (nA)
        stdev: float
            Standard deviation of the OU process (nA)
        delay: float
            The delay of current onset from the start of the simulation (ms)
        dur: float
            The duration of the current input (ms)
        seed: int, optional
            Seed for the random number generator
        """
        seed = np.random.randint(1e16) if seed is None else seed
        loc = MorphLoc(loc, self)
        # create the current clamp
        if tau > 1e-9:
            iclamp = h.OUClamp(self.sections[loc['node']](loc['x']))
            iclamp.tau = tau
        else:
            iclamp = h.WNclamp(self.sections[loc['node']](loc['x']))
        iclamp.mean = mean # nA
        iclamp.stdev = stdev # nA
        iclamp.delay = delay + self.t_calibrate # ms
        iclamp.dur = dur # ms
        iclamp.seed_usr = seed # ms
        iclamp.dt_usr = self.dt # ms
        # store the iclamp
        self.iclamps.append(iclamp)

    def addOUconductance(self, loc, tau, mean, stdev, e_r, delay, dur, seed=None):
        """
        Injects a Ornstein-Uhlenbeck conductance at a given location

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the conductance.
        tau: float
            Time-scale of the OU process (ms)
        mean: float
            Mean of the OU process (uS)
        stdev: float
            Standard deviation of the OU process (uS)
        e_r: float
            Reversal of the current (mV)
        delay: float
            The delay of current onset from the start of the simulation (ms)
        dur: float
            The duration of the current input (ms)
        seed: int, optional
            Seed for the random number generator
        """
        seed = np.random.randint(1e16) if seed is None else seed
        loc = MorphLoc(loc, self)
        # create the current clamp
        iclamp = h.OUConductance(self.sections[loc['node']](loc['x']))
        iclamp.tau = tau
        iclamp.mean = mean # uS
        iclamp.stdev = stdev # uS
        iclamp.e_r = e_r # mV
        iclamp.delay = delay + self.t_calibrate # ms
        iclamp.dur = dur # ms
        iclamp.seed_usr = seed # ms
        iclamp.dt_usr = self.dt # ms
        # store the iclamp
        self.iclamps.append(iclamp)

    def addOUReversal(self, loc, tau, mean, stdev, g_val, delay, dur, seed=None):
        seed = np.random.randint(1e16) if seed is None else seed
        loc = MorphLoc(loc, self)
        # create the current clamp
        iclamp = h.OUReversal(self.sections[loc['node']](loc['x']))
        iclamp.tau = tau # ms
        iclamp.mean = mean # mV
        iclamp.stdev = stdev # mV
        iclamp.g = g_val # uS
        iclamp.delay = delay + self.t_calibrate # ms
        iclamp.dur = dur # ms
        iclamp.seed_usr = seed # ms
        iclamp.dt_usr = self.dt # ms
        # store the iclamp
        self.iclamps.append(iclamp)

    def addVClamp(self, loc, e_c, dur):
        """
        Adds a voltage clamp at a given location

        Parameters
        ----------
        loc: dict, tuple or :class:`neat.MorphLoc`
            The location of the conductance.
        e_c: float
            The clamping voltage (mV)
        dur: float, ms
            The duration of the voltage clamp
        """
        loc = MorphLoc(loc, self)
        # add the voltage clamp
        vclamp = h.SEClamp(self.sections[loc['node']](loc['x']))
        vclamp.rs = 0.01
        vclamp.dur1 = dur
        vclamp.amp1 = e_c
        # store the vclamp
        self.vclamps.append(vclamp)

    # def addRecLoc(self, loc):
    #     self.addLoc(loc, 'rec locs')

    def setSpikeTrain(self, syn_index, syn_weight, spike_times):
        """
        Each hoc point process that receive spikes through should by appended to
        the synapse stack (stored under the list `self.syns`).

        Default :class:`NeuronSimTree` point processes that are added to
        `self.syns` are:
        - `self.addDoubleExpCurrent()`
        - `self.addExpSyn()`
        - `self.addDoubleExpSyn()`
        - `self.addDoubleExpSyn()`
        - `self.addNMDASynapse()`
        - `self.addDoubleExpNMDASynapse()`

        With this function, these synapse can be set to receive a specific spike
        train.

        Parameters
        ----------
        syn_index: int
            index of the point process in the synapse stack
        syn_weight: float
            weight of the synapse (maximal value of the conductance window)
        spike_times: list or `np.array` of floats
            the spike times
        """
        # add spiketrain
        spks = np.array(spike_times) + self.t_calibrate
        spks_vec = h.Vector(spks.tolist())
        vecstim = h.VecStim()
        vecstim.play(spks_vec)
        netcon = h.NetCon(vecstim, self.syns[syn_index], 0, self.dt, syn_weight)
        # store the objects
        self.vecs.append(spks_vec)
        self.vecstims.append(vecstim)
        self.netcons.append(netcon)

    def run(self, t_max, downsample=1,
            record_from_syns=False, record_from_iclamps=False, record_from_vclamps=False,
            record_from_channels=False, record_v_deriv=False, record_concentrations=[],
            pprint=False):
        """
        Run the NEURON simulation. Records at all locations stored
        under the name 'rec locs' on `self` (see `MorphTree.storeLocs()`)

        Parameters
        ----------
        t_max: float
            Duration of the simulation
        downsample: int (> 0)
            Records the state of the model every `downsample` time-steps
        record_from_syns: bool (default ``False``)
            Record currents of synapstic point processes (in `self.syns`).
            Accessible as `np.ndarray` in the output dict under key 'i_syn'
        record_from_iclamps: bool (default ``False``)
            Record currents of iclamps (in `self.iclamps`)
            Accessible as `np.ndarray` in the output dict under key 'i_clamp'
        record_from_vclamps: bool (default ``False``)
            Record currents of vclamps (in `self.vclamps`)
            Accessible as `np.ndarray` in the output dict under key 'i_vclamp'
        record_from_channels: bool (default ``False``)
            Record channel state variables from `neat` defined channels in `self`,
            at locations stored under 'rec locs'
            Accessible as `np.ndarray` in the output dict under key 'chan'
        record_v_deriv: bool (default ``False``)
            Record voltage derivative at locations stored under 'rec locs'
            Accessible as `np.ndarray` in the output dict under key 'dv_dt'
        record_from_concentrations: bool (default ``False``)
            Record ion concentration at locations stored under 'rec locs'
            Accessible as `np.ndarray` in the output dict with as key the ion's
            name

        Returns
        -------
        dict
            Dictionary with the results of the simulation. Contains time and
            voltage as `np.ndarray` at locations stored under the name '
            rec locs', respectively with keys 't' and 'v_m'. Also contains
            traces of other recorded variables if the option to record them was
            set to ``True``
        """
        assert isinstance(downsample, int) and downsample > 0
        # simulation time recorder
        res = {'t': h.Vector()}
        res['t'].record(h._ref_t)
        # voltage recorders
        res['v_m'] = []
        for loc in self.getLocs('rec locs'):
            res['v_m'].append(h.Vector())
            res['v_m'][-1].record(self.sections[loc['node']](loc['x'])._ref_v)
        # synapse current recorders
        if record_from_syns:
            res['i_syn'] = []
            for syn in self.syns:
                res['i_syn'].append(h.Vector())
                res['i_syn'][-1].record(syn._ref_i)
        # current clamp current recorders
        if record_from_iclamps:
            res['i_clamp'] = []
            for iclamp in self.iclamps:
                res['i_clamp'].append(h.Vector())
                res['i_clamp'][-1].record(iclamp._ref_i)
        # voltage clamp current recorders
        if record_from_vclamps:
            res['i_vclamp'] = []
            for vclamp in self.vclamps:
                res['i_vclamp'].append(h.Vector())
                res['i_vclamp'][-1].record(vclamp._ref_i)
        # channel state variable recordings
        if record_from_channels:
            res['chan'] = {}
            channel_names = self.getChannelsInTree()
            for channel_name in channel_names:
                res['chan'][channel_name] = {str(var): [] for var in self.channel_storage[channel_name].statevars}
                for loc in self.getLocs('rec locs'):
                    for ind, varname in enumerate(self.channel_storage[channel_name].statevars):
                        var = str(varname)
                        # assure xcoordinate is refering to proper neuron section (not endpoint)
                        xx = loc['x']
                        if xx < 1e-3: xx += 1e-3
                        elif xx > 1. - 1e-3: xx -= 1e-3
                        # create the recorder
                        try:
                            rec = h.Vector()
                            exec('rec.record(self.sections[loc[0]](xx).' + mechname[channel_name] + '._ref_' + str(var) +')')
                            res['chan'][channel_name][var].append(rec)
                        except AttributeError:
                            # the channel does not exist here
                            res['chan'][channel_name][var].append([])
        if len(record_concentrations) > 0:
            for c_ion in record_concentrations:
                res[c_ion] = []
                for loc in self.getLocs('rec locs'):
                    res[c_ion].append(h.Vector())
                    exec('res[c_ion][-1].record(self.sections[loc[\'node\']](loc[\'x\'])._ref_' + c_ion + 'i)')
        # record voltage derivative
        if record_v_deriv:
            res['dv_dt'] = []
            for ii, loc in enumerate(self.getLocs('rec locs')):
                res['dv_dt'].append(h.Vector())
                # res['dv_dt'][-1].deriv(res['v_m'][ii], self.dt)

        # initialize
        # neuron.celsius=37.
        h.finitialize(self.v_init)
        h.dt = self.dt

        # simulate
        if pprint: print('>>> Simulating the NEURON model for ' + str(t_max) + ' ms. <<<')
        start = posix.times()[0]
        neuron.run(t_max + self.t_calibrate)
        stop = posix.times()[0]
        if pprint: print('>>> Elapsed time: ' + str(stop-start) + ' seconds. <<<')
        runtime = stop-start

        # compute derivative
        if 'dv_dt' in res:
            for ii, loc in enumerate(self.getLocs('rec locs')):
                res['dv_dt'][ii].deriv(res['v_m'][ii], h.dt, 2)
                res['dv_dt'][ii] = np.array(res['dv_dt'][ii])[self.indstart:][::downsample]
            res['dv_dt'] = np.array(res['dv_dt'])
        # cast recordings into numpy arrays
        res['t'] = np.array(res['t'])[self.indstart:][::downsample] - self.t_calibrate
        for key in set(res.keys()) - {'t', 'chan', 'dv_dt'}:
            if key in res and len(res[key]) > 0:
                res[key] = np.array([np.array(reslist)[self.indstart:][::downsample] \
                                     for reslist in res[key]])
                if key in ('i_syn', 'i_clamp', 'i_vclamp'):
                    res[key] *= -1.
        # cast channel recordings into numpy arrays
        if 'chan' in res:
            for channel_name in channel_names:
                channel = self.channel_storage[channel_name]
                for ind0, varname in enumerate(channel.statevars):
                    var = str(varname)
                    for ind1 in range(len(self.getLocs('rec locs'))):
                        res['chan'][channel_name][var][ind1] = \
                                np.array(res['chan'][channel_name][var][ind1])[self.indstart:][::downsample]
                        if len(res['chan'][channel_name][var][ind1]) == 0:
                            res['chan'][channel_name][var][ind1] = np.zeros_like(res['t'])
                    res['chan'][channel_name][var] = \
                            np.array(res['chan'][channel_name][var])
                # compute P_open
                # sv = np.zeros((len(channel.statevars), len(self.getLocs('rec locs')), len(res['t'])))
                sv = {}
                for varname in channel.statevars:
                    var = str(varname)
                    sv[var] = res['chan'][channel_name][var]
                res['chan'][channel_name]['p_open'] = channel.computePOpen(res['v_m'], **sv)

        return res

    def calcEEq(self, t_dur=100., set_e_eq=True):
        """
        Compute the equilibrium potentials in the middle (``x=0.5``) of each node.

        Parameters
        ----------
        t_dur: float (optional, default ``100.`` ms)
            The duration of the simulation
        set_e_eq: bool (optional, default ``True``)
            Store the equilibrium potential as the ``PhysNode.e_eq`` attribute
        """
        self.initModel(dt=self.dt, t_calibrate=self.t_calibrate,
                       v_init=self.v_init, factor_lambda=self.factor_lambda)
        self.storeLocs([(n.index, 0.5) for n in self], name='rec locs')
        res = self.run(t_dur)
        v_eq = res['v_m'][:-1]
        if set_e_eq:
            for (node, e) in zip(self, v_eq): node.setEEq(e_eq)


        return v_eq

    def calcImpedanceMatrix(self, locarg, i_amp=0.001, t_dur=100., pplot=False):
        if isinstance(locarg, list):
            locs = [MorphLoc(loc, self) for loc in locarg]
        elif isinstance(locarg, str):
            locs = self.getLocs(locarg)
        else:
            raise IOError('`locarg` should be list of locs or string')
        z_mat = np.zeros((len(locs), len(locs)))
        for ii, loc0 in enumerate(locs):
            for jj, loc1 in enumerate(locs):
                self.initModel(dt=self.dt, t_calibrate=self.t_calibrate,
                               v_init=self.v_init, factor_lambda=self.factor_lambda)
                self.addIClamp(loc0, i_amp, 0., t_dur)
                self.storeLocs([loc0, loc1], 'rec locs', warn=False)
                # simulate
                res = self.run(t_dur)
                # voltage deflections
                # v_trans = res['v_m'][1][-int(1./self.dt)] - self[loc1['node']].e_eq
                v_trans = res['v_m'][1][-int(1./self.dt)] - res['v_m'][1][0]
                # compute impedances
                z_mat[ii, jj] = v_trans / i_amp
                if pplot:
                    import matplotlib.pyplot as pl
                    pl.figure()
                    pl.plot(res['t'], res['v_m'][1])
                    pl.show()

        return z_mat

    def calcImpedanceKernelMatrix(self, locarg, i_amp=0.001,
                                                dt_pulse=0.1, t_max=100.):
        tk = np.arange(0., t_max, self.dt)
        if isinstance(locarg, list):
            locs = [MorphLoc(loc, self) for loc in locarg]
        elif isinstance(locarg, str):
            locs = self.getLocs(locarg)
        else:
            raise IOError('`locarg` should be list of locs or string')
        zk_mat = np.zeros((len(tk), len(locs), len(locs)))
        for ii, loc0 in enumerate(locs):
            for jj, loc1 in enumerate(locs):
                loc1 = locs[jj]
                self.initModel(dt=self.dt, t_calibrate=self.t_calibrate,
                               v_init=self.v_init, factor_lambda=self.factor_lambda)
                self.addIClamp(loc0, i_amp, 0., dt_pulse)
                self.storeLocs([loc0, loc1], 'rec locs', warn=False)
                # simulate
                res = self.run(t_max)
                # voltage deflections
                v_trans = res['v_m'][1][1:] - self[loc1['node']].e_eq
                # compute impedances
                zk_mat[:, ii, jj] = v_trans / (i_amp * dt_pulse)
        return tk, zk_mat


class NeuronCompartmentNode(NeuronSimNode):
    def __init__(self, index):
        super().__init__(index)

    def getChildNodes(self, skip_inds=[]):
        return super().getChildNodes(skip_inds=skip_inds)

    def _makeSection(self, pprint=False):
        compartment = neuron.h.Section(name=str(self.index))
        compartment.push()
        # create the compartment
        if 'points_3d' in self.content:
            points = self.content['points_3d']
            h.pt3dadd(*points[0], sec=compartment)
            h.pt3dadd(*points[1], sec=compartment)
            h.pt3dadd(*points[2], sec=compartment)
            h.pt3dadd(*points[3], sec=compartment)
        else:
            compartment.diam = 2. * self.R  # section radius [um] (NEURON takes diam = 2*r)
            compartment.L = self.L # section length [um]
        # set number of segments to one
        compartment.nseg = 1

        # set parameters
        compartment.cm = self.c_m # uF/cm^2
        compartment.Ra = self.r_a*1e6 # MOhm*cm --> Ohm*cm
        # insert membrane currents
        for key, current in self.currents.items():
            if current[0] > 1e-10:
                compartment.insert(mechname[key])
                for seg in compartment:
                    exec('seg.' + mechname[key] + '.g = ' + str(current[0]) + '*1e-6') # uS/cm^2 --> S/cm^2
                    exec('seg.' + mechname[key] + '.e = ' + str(current[1])) # mV
        # insert concentration mechanisms
        for ion, params in self.concmechs.items():
            compartment.insert(mechname[ion])
            for seg in compartment:
                for param, value in params.items():
                    exec('seg.' + mechname[ion] + '.' + param + ' = ' + str(value))
        h.pop_section()

        if pprint:
            print(self)
            print(('>>> compartment length = %.2f um'%compartment.L))
            print(('>>> compartment diam = %.2f um'%compartment.diam))
            print(('>>> compartment nseg = ' + str(compartment.nseg)))

        return compartment


class NeuronCompartmentTree(NeuronSimTree):
    def __init__(self, t_calibrate=0., dt=0.025, v_init=-75.):
        super().__init__(file_n=None, types=[1,3,4],
                         t_calibrate=t_calibrate, dt=dt, v_init=v_init)

    # redefinition of bunch of standard functions to not include skip inds by default
    def __getitem__(self, index, skip_inds=[]):
        return super().__getitem__(index, skip_inds=skip_inds)

    def getNodes(self, recompute_flag=0, skip_inds=[]):
        return super().getNodes(recompute_flag=recompute_flag, skip_inds=skip_inds)

    def __iter__(self, node=None, skip_inds=[]):
        return super().__iter__(node=node, skip_inds=skip_inds)

    def _findNode(self, node, index, skip_inds=[]):
        return super()._findNode(node, index, skip_inds=skip_inds)

    def _gatherNodes(self, node, node_list=[], skip_inds=[]):
        return super()._gatherNodes(node, node_list=node_list, skip_inds=skip_inds)

    def _createCorrespondingNode(self, node_index):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
            node_index: int
                index of the new node
        """
        return NeuronCompartmentNode(node_index)

    def _createNeuronTree(self, pprint):
        for node in self:
            # create the NEURON section
            compartment = node._makeSection(pprint=pprint)
            # connect with parent section
            if not self.isRoot(node):
                compartment.connect(self.sections[node.parent_node.index], 0.5, 0)
            # store
            self.sections.update({node.index: compartment})
            # create a static shunt
            shunt = node._makeShunt(compartment)
            if shunt is not None:
                self.shunts.append(shunt)


def createReducedNeuronModel(ctree, fake_c_m=1., fake_r_a=100.*1e-6, method=2):
    # calculate geometry that will lead to correct constants
    arg1, arg2 = ctree.computeFakeGeometry(fake_c_m=fake_c_m, fake_r_a=fake_r_a,
                                                 factor_r_a=1e-6, delta=1e-10,
                                                 method=method)
    if method == 1:
        points = arg1; surfaces = arg2
        sim_tree = ctree.__copy__(new_tree=NeuronCompartmentTree())
        for ii, comp_node in enumerate(ctree):
            pts = points[ii]
            sim_node = sim_tree.__getitem__(comp_node.index, skip_inds=[])
            sim_node.setP3D(np.array(pts[0][:3]), (pts[0][3] + pts[-1][3]) / 2., 3)

        # fill the tree with the currents
        for ii, sim_node in enumerate(sim_tree):
            comp_node = ctree[ii]
            sim_node.currents = {chan: [g / surfaces[comp_node.index], e] \
                                         for chan, (g, e) in comp_node.currents.items()}
            sim_node.concmechs = copy.deepcopy(comp_node.concmechs)
            for concmech in sim_node.concmechs.values():
                concmech.gamma *= surfaces[comp_node.index]
            sim_node.c_m = fake_c_m
            sim_node.r_a = fake_r_a
            sim_node.content['points_3d'] = points[comp_node.index]
    elif method == 2:
        lengths = arg1 ; radii = arg2
        surfaces = 2. * np.pi * radii * lengths
        sim_tree = ctree.__copy__(new_tree=NeuronCompartmentTree())
        for ii, comp_node in enumerate(ctree):
            sim_node = sim_tree.__getitem__(comp_node.index, skip_inds=[])
            if sim_tree.isRoot(sim_node):
                sim_node.setP3D(np.array([0.,0.,0.]), radii[ii]*1e4, 1)
            else:
                sim_node.setP3D(np.array([sim_node.parent_node.xyz[0]+lengths[ii]*1e4, 0., 0.]),
                                 radii[ii]*1e4, 3)

        # fill the tree with the currents
        for ii, sim_node in enumerate(sim_tree):
            comp_node = ctree[ii]
            sim_node.currents = {chan: [g / surfaces[comp_node.index], e] \
                                         for chan, (g, e) in comp_node.currents.items()}
            sim_node.concmechs = copy.deepcopy(comp_node.concmechs)
            for concmech in sim_node.concmechs.values():
                concmech.gamma *= surfaces[comp_node.index]
            sim_node.c_m = fake_c_m
            sim_node.r_a = fake_r_a
            sim_node.R = radii[comp_node.index]*1e4    # convert to [um]
            sim_node.L = lengths[comp_node.index]*1e4  # convert to [um]
    return sim_tree







"""
File contains:

    - `neat.SNode`
    - `neat.STree`

Authors: B. Torben-Nielsen (legacy code), W. Wybo
"""

import numpy as np

import warnings
import copy
from collections import Counter
from functools import reduce

class SNode(object):
    """
    Simple Node for use with a simple Tree (`neat.STree`)

    Parameters
    ----------
    index: int
        index of the node

    Attributes
    ----------
    index: int
        index of the node
    parent_node: `neat.SNode` or ``None``
        parent of node, ``None`` means node is root
    child_nodes: list of `neat.SNode`
        child nodes of ``self``, empty list means node is leaf
    content: dict
        arbitrary items can be stored at the node
    """

    def __init__(self, index):
        self.index = index
        self._parent_node = None
        self._child_nodes = []
        self._content = {}

    def getParentNode(self):
        return self._parent_node

    def setParentNode(self, parent_node):
        self._parent_node = parent_node

    parent_node = property(getParentNode, setParentNode)

    def getChildNodes(self, **kwargs):
        return self._child_nodes

    def setChildNodes(self, cnodes):
        self._child_nodes = cnodes

    child_nodes = property(getChildNodes, setChildNodes)

    def addChild(self, child_node):
        self._child_nodes.append(child_node)

    def getContent(self):
        return self._content

    def setContent(self, content):
        if isinstance(content, dict):
            self._content = content
        else :
            raise Exception("SNode.setContent must receive a dict")

    content = property(getContent, setContent)

    def makeEmpty(self):
        '''
        Remove content and references to parent and child nodes
        '''
        self._parent_node = None
        self._content = None
        self._child_nodes = []

    def removeChild(self, child_node):
        '''
        Remove a single child node

        Parameters
        ----------
        child_node: `neat.SNode`
            child node to be removed
        '''
        self._child_nodes.remove(child_node)

    def __getitem__(self, key):
        return self._content[key]

    def __setitem__(self, key, value):
        self._content[key] = value

    def __str__(self, with_parent=False, with_children=False):
        node_string = 'SNode ' + str(self.index)
        if with_parent:
            node_string += ', Parent: ' + str(self.parent_node)
        if with_children:
            node_string += ', Children:' + \
                            str([str(cnode) for cnode in self.child_nodes])
        return node_string

    def __copy__(self, new_node=None):
        """
        experimental, untested
        """
        if new_node is None:
            # new_node = self.__class__(self.index)
            for child in self._child_nodes :
                ret.addChild(child)
            try:
                ret.content = self.content
            except AttributeError:
                # no content variable set
                pass
            ret.setParentNode(self._parent_node)
        else:
            orig_keys = set(self.__dict__.keys()) - {'_parent_node', '_child_nodes'}
            copy_keys = orig_keys.intersection(set(new_node.__dict__.keys()))
            for key in copy_keys:
                new_node.__dict__[key] = copy.deepcopy(self.__dict__[key])
            return new_node


class STree(object):
    """
    A simple tree for use with a simple Node (`neat.SNode`).

    Generic implementation of a tree structure as a linked list extended with
    some convenience functions

    Parameters
    ----------
    root: `neat.SNode`, optional
        The root of the tree, default is ``None`` which creates an empty tree

    Attributes
    ----------
    root: `neat.SNode`
        The root of the tree
    """

    def __init__(self, root=None):
        """
        Initialize an empty tree by default
        """
        if root is not None:
            self.root = root
        else:
            self._root = None

    def __getitem__(self, index, **kwargs):
        """
        Returns the node with given index, if no such node is in the tree, None
        is returned.

        Parameters
        ----------
            index: int
                the index of the node to be found

        Returns
        -------
            `neat.SNode` or None
        """
        return self._findNode(self.root, index)

    def _findNode(self, node, index):
        """
        Sweet breadth-first/stack iteration to replace the recursive call.
        Traverses the tree until it finds the node you are looking for.
        Returns SNode when found and None when not found

        Parameters
        ----------
            node: `neat.SNode` (optional)
                node where the search is started
            index: int
                the index of the node to be found

        Returns
        -------
            `neat.SNode`
        """
        stack = [];
        stack.append(node)
        while len(stack) != 0:
            for cnode in stack:
                if cnode.index == index:
                    return cnode
                else:
                    stack.remove(cnode)
                    stack.extend(cnode.getChildNodes())
        return None # Not found!

    def __len__(self, node=None):
        """
        Return the number of nodes in the tree. If an input node is specified,
        the number of nodes in the subtree of the input node is returned

        Parameters
        ----------
            node: `neat.SNode` (optional)
                The starting node. Defaults to root

        Returns
        -------
            int
        """
        self._node_count = 0
        for node in self.__iter__(node=node):
            self._node_count += 1
        return self._node_count

    def __iter__(self, node=None, **kwargs):
        """
        Iterate over the nodes in the subtree of the given node.

        Beware, if the given node is not in the tree, it will simply iterate
        over the subtree of the given node.

        Parameters
        ----------
            node: `neat.SNode` (optional)
                The starting node. Defaults to the root
        """
        if node is None:
            node = self.root
        if node is not None:
            yield node
            if node is not None:
                for cnode in node.getChildNodes():
                    for inode in self.__iter__(cnode):
                        yield inode

    def __str__(self, node=None):
        """
        Generate a string of the subtree of the given node.

        Beware, if the given node is not in the tree, it will simply iterate
        over the subtree of the given node.

        Parameters
        ----------
            node: `neat.SNode` (optional)
                The starting node. Defaults to the root
        """
        if node is None:
            node = self.root
        tree_string = '>>> Tree'
        for iternode in self.__iter__(node):
            tree_string += '\n    ' + iternode.__str__(with_parent=True)
        return tree_string

    def checkOrdered(self):
        """
        Check if the indices of the tree are number in the same order as they
        appear in the iterator
        """
        return list(range(len(self))) == [node.index for node in self]

    def getNodes(self, recompute_flag=1):
        """
        Build a list of all the nodes in the tree

        Parameters
        ----------
            recompute_flag: bool
                whether or not to re-evaluate the node list

        Returns
        -------
            list of :class:`Snode`
        """
        if not hasattr(self, '_nodes') or recompute_flag:
            self._nodes = []
            self._gatherNodes(self.root, self._nodes)
        return self._nodes

    def setNodes(self, illegal):
        raise AttributeError("`nodes` is a read-only attribute")

    nodes = property(getNodes, setNodes)

    def gatherNodes(self, node):
        """
        Build a list of all the nodes in the subtree of the provided node

        Parameters
        ----------
            node: :class:`Snode`
                starting point node

        Returns
        -------
            list of :class:`Snode`
        """
        nodes = []
        self._gatherNodes(node, nodes)
        return nodes

    def _gatherNodes(self, node, node_list=[]):
        """
        Append node to list and recurse to its child nodes

        Parameters
        ----------
            node: `neat.SNode`
            node_list: list of `neat.SNode`
        """
        node_list.append(node)
        for cnode in node.child_nodes:
            self._gatherNodes(cnode, node_list=node_list)

    def getLeafs(self, recompute_flag=1):
        """
        Get all leaf nodes in the tree.

        Parameters
        ----------
            recompute_flag: bool
                Whether to force recomputing the leaf list. Defaults to 1.
        """
        if not hasattr(self, '_leafs') or recompute_flag:
            self._leafs = [node for node in self if self.isLeaf(node)]
        return self._leafs

    def setLeafs(self, illegal):
        raise AttributeError("`leafs` is a read-only attribute")

    leafs = property(getLeafs, setLeafs)

    def setRoot(self, node):
        """
        Set the root node of the tree

        Parameters
        ----------
            node: `neat.SNode`
                node to be set as root of the tree
        """
        node.parent_node = None
        self._root = node

    def getRoot(self):
        return self._root

    root = property(getRoot, setRoot)

    def isRoot(self, node):
        """
        Check if input node is root of the tree.

        Parameters
        ----------
            node: `neat.SNode`
        """
        if node.getParentNode() is not None:
            return False
        else:
            return True

    def isLeaf(self, node):
        """
        Check if input node is a leaf of the tree

        Parameters
        ----------
            node: `neat.SNode`
        """
        if len(node.getChildNodes()) == 0:
            return True
        else:
            return False

    def _createCorrespondingNode(self, node_index):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
            node_index: int
                index of the new node
        """
        return SNode(node_index)

    def addNodeWithParentFromIndex(self, node_index, pnode, *args, **kwargs):
        """
        Create a node with the given index and add it to the tree under a
        specific parent node.

        Parameters
        ----------
            node_index: int
                index of the new node
            pnode: `neat.SNode`
                parent node of the newly added node

        Raises
        ------
            ValueError
                if ``node_index`` is already in the tree
        """
        if self[node_index] == None:
            node = self._createCorrespondingNode(node_index, *args, **kwargs)
            self.addNodeWithParent(node, pnode)
        else:
            raise ValueError('Index %d is already exists in the tree.')

    def addNodeWithParent(self, node, pnode):
        """
        Add a node to the tree under a specific parent node

        Parameters
        ----------
            node: `neat.SNode`
                node to be added
            pnode: `neat.SNode`
                parent node of the newly added node
        """
        if pnode is not None:
            node.setParentNode(pnode)
            pnode.addChild(node)
        else:
            warnings.warn('`pnode` was `None`, did nothing.')

    def softRemoveNode(self, node):
        """
        Remove a node and its subtree from the tree by deleting the reference
        to it in its parent. Internally, the node and its linked subtree are not
        changed

        Parameters
        ----------
            node: `neat.SNode`
                node to be removed
        """
        node.getParentNode().removeChild(node)

    def removeNode(self, node):
        """
        Remove a node as well as its subtree from the tree

        Parameters
        ----------
            node: `neat.SNode`
                node to be removed
        """
        node.getParentNode().removeChild(node)
        self._deepRemove(node)

    def _deepRemove(self, node):
        cnodes = node.getChildNodes()
        node.makeEmpty()
        for cnode in cnodes:
            self._deepRemove(cnode)

    def removeSingleNode(self, node):
        """
        Remove a single node from the tree. The nodes' children become the
        children of the nodes' parent.

        Parameters
        ----------
            node: `neat.SNode`
                node to be removed
        """
        if node == self.root:
            raise ValueError('Removing root is forbidden')
        cnodes = node.getChildNodes()
        pnode = node.getParentNode()
        pnode.removeChild(node)
        for cnode in cnodes:
            cnode.setParentNode(pnode)
            pnode.addChild(cnode)

    def insertNode(self, node, pnode, pcnodes=[]):
        """
        Insert a node in the tree as a child of a specified parent. The
        original children of the parent that will become children of the node
        are specified in the ``pcnodes`` list

        Parameters
        ----------
            node: `neat.SNode`
                the node that is to be inserted
            pnode: `neat.SNode`
                the node that will become parent of the node that is to be
                inserted
            pcnodes: list of `neat.SNode`
                the current children of the pnode that will become children of
                the node
        """
        if pnode != None:
            cnodes = pnode.getChildNodes()
            for pcnode in pcnodes:
                if pcnode in cnodes:
                    ind = cnodes.index(pcnode)
                    del cnodes[ind]
                    node.addChild(pcnode)
                    pcnode.setParentNode(node)
                else:
                    warnings.warn(str(pcnode) + ' is not a child of ' \
                                              + str(pnode) + ', ignoring it')
            node.setParentNode(pnode)
            pnode.addChild(node)
        if pnode == None:
            cnode = self.root
            cnode.setParentNode(node)
            node.setParentNode(None)
            node.addChild(cnode)
            self.root = node

    def resetIndices(self, n=0):
        """
        Resets the indices in the order they appear in a depth-first iteration
        """
        for ind, node in enumerate(self):
            node.index = ind+n

    def getSubTree(self, node, new_tree=None):
        """
        Get the subtree of the specified node. The root of the subtree is a new
        node with the same children as the original node, but None instead of a
        parent.

        Parameters
        ----------
            node: `neat.SNode`
                root of the sub tree

        Returns
        -------
        `neat.STree`
            Subtree of with ``node`` as root
        """
        if new_tree is None:
            new_tree = STree()

        new_node = new_tree._createCorrespondingNode(node.index)
        node.__copy__(new_node=new_node)
        new_node.setParentNode(None)
        new_tree.setRoot(new_node)

        self._recurseCopy(node, new_tree)

        return new_tree

    def depthOfNode(self, node):
        """
        compute the depth of the node (number of edges between node and root)

        Parameters
        ----------
        node: `neat.SNode`

        Returns
        -------
        int
            depth of the node
        """
        return len(self.pathToRoot(node)) - 1

    def degreeOfNode(self, node):
        """
        Compute the degree (number of leafs in its subtree) of a node.

        Parameters
        ----------
            node: `neat.SNode`
        """
        return len([node for node in self.__iter__(node) if self.isLeaf(node)])

    def orderOfNode(self, node):
        """
        Compute the order (number of bifurcations from the root) of a node.

        Parameters
        ----------
            node: `neat.SNode`
        """
        ptr = self.pathToRoot(node)
        order = 0
        for node in ptr:
            if len(node.child_nodes) > 1:
                order += 1
        # order is on [0,max_order] thus subtract 1 from this calculation
        return order - 1

    def pathToRoot(self, node):
        """
        Return the path from a given node to the root

        Parameters:
            node: `neat.SNode`

        Returns
        -------
            list of `neat.SNode`
                List of nodes from ``node`` to root. First node is the input node
                and last node is the root
        """
        nodes = []
        self._goUpFrom(node, nodes)
        return nodes

    def _goUpFrom(self, node, nodes):
        nodes.append(node)
        pnode = node.getParentNode()
        if pnode != None :
            self._goUpFrom(pnode, nodes)

    def pathBetweenNodes(self, from_node, to_node):
        """
        Inclusive path from ``from_node`` to ``to_node``.

        Parameters
        ----------
            from_node: `neat.SNode`
            to_node: `neat.SNode`

        Returns
        -------
            list of `neat.SNode`
                List of nodes representing the direct path between ``from_node``
                and ``to_node``, which are respectively the first and last nodes
                in the list.
        """
        path1 = self.pathToRoot(from_node)[::-1]
        path2 = self.pathToRoot(to_node)[::-1]
        path = path1 if len(path1) < len(path2) else path2
        ind = next((ii for ii in range(len(path)) if path1[ii] != path2[ii]),
                   len(path))
        return path1[ind:][::-1] + path2[ind-1:]

    def pathBetweenNodesDepthFirst(self, from_node, to_node):
        """
        Inclusive path from ``from_node`` to ``to_node``, ginven in a depth-
        first ordering.

        Parameters
        ----------
            from_node: `neat.SNode`
            to_node: `neat.SNode`

        Returns
        -------
            list of `neat.SNode`
                List of nodes representing the direct path between ``from_node``
                and ``to_node``, which are respectively the first and last nodes
                in the list.
        """
        path1 = self.pathToRoot(from_node)[::-1]
        path2 = self.pathToRoot(to_node)[::-1]
        path = path1 if len(path1) < len(path2) else path2
        ind = next((ii for ii in range(len(path)) if path1[ii] != path2[ii]),
                   len(path))
        return path1[ind-1:] + path2[ind:]

    def getNodesInSubtree(self, ref_node, subtree_root=None):
        """
        Returns the nodes in the subtree that contains the given reference nodes
        and has the given subtree root as root. If the subtree root is not
        provided, the subtree of the first child node of the root on the path to
        the reference node is given (plus the root)

        Parameters
        ----------
            ref_node: `neat.SNode`
                the reference node that is in the subtree
            subtree_root: `neat.SNode`
                what is to be the root of the subtree. If this node is not on
                the path from reference node to root, a ValueError is raised

        Returns
        -------
            list of `neat.SNode`
                List of all nodes in the subtree. It's root is in the first
                position
        """
        if subtree_root == None:
            subtree_root = self.root
        ref_path = self.pathBetweenNodes(ref_node, subtree_root)
        if subtree_root in ref_path:
            if len(ref_path) > 1:
                subtree_nodes = [subtree_root] \
                                + self.gatherNodes(ref_path[-2])
            else:
                subtree_nodes = [ref_node] # both input nodes are the same
        else:
            raise ValueError('|subtree_root| not in path from |ref_node| \
                                root')
        return subtree_nodes

    def sisterLeafs(self, node):
        """
        Find the leafs that are in the subtree of the nearest bifurcation node
        up from the input node.

        Parameters
        ----------
            node: `neat.SNode`
                Starting node for search

        Returns
        -------
        node: `neat.SNode`
            the bifurcation node
        sister_leafs: list of `neat.SNode`
            The first element is the input node. The others are the leafs
            of the subtree emanating from the bifurcation node that are not
            in the subtree from the input node.
        corresponding_children: list of `neat.SNode`
            The children of the bifurcation node. If the number of leafs
            ``sister_leafs`` is the same as the number of
            ``corresponding_children``, the subtree of each element of
            ``corresponding_children`` has exactly one leaf, the corresponding
            element in ``sister_leafs``
        """
        sleafs = [node]; cchildren = []
        snode = self._goUpUntil(node, None, sl=sleafs, cc=cchildren)
        return snode, sleafs, cchildren

    def _goUpUntil(self, node, c_node, sl=[], cc=[]):
        c_nodes = node.getChildNodes()
        if (c_node != None and len(c_nodes) > 1) or self.isRoot(node):
            cc.append(c_node)
            for c_node_ in set(c_nodes) - {c_node}:
                self._goDownUntil(c_node_, sl=sl)
                cc.append(c_node_)
        else:
            p_node = node.getParentNode()
            node = self._goUpUntil(p_node, node, sl=sl, cc=cc)
        return node

    def _goDownUntil(self, node, sl=[]):
        c_nodes = node.getChildNodes()
        if len(c_nodes) == 0:
            sl.append(node)
        else:
            for c_node in c_nodes:
                self._goDownUntil(c_node, sl=sl)

    def upBifurcationNode(self, node, cnode=None):
        """
        Find the nearest bifurcation node up (towards root) from the input node.

        Parameters
        ----------
        node: `neat.SNode`
            Starting node for search
        cnode: `neat.SNode`
            For recursion, don't change default

        Returns
        -------
        node: `neat.SNode`
            the bifurcation node
        cnode: `neat.SNode`
            The bifurcation node's child on the path to the input node.

        """
        if cnode == None or len(node.getChildNodes()) <= 1:
            pnode = node.getParentNode()
            if pnode != None:
                node, cnode = self.upBifurcationNode(pnode, cnode=node)
        return node, cnode

    def downBifurcationNode(self, node):
        """
        Find the nearest bifurcation node down (towards leafs) from the input node.

        Parameters
        ----------
        node: `neat.SNode`
            Starting node for search

        Returns
        -------
        node: `neat.SNode`
            the bifurcation node
        """
        if len(node.child_nodes) > 1:
            return node
        elif len(node.child_nodes) == 0:
            return None
        else:
            self.downBifurcationNode(node.child_nodes[0])

    def getBifurcationNodes(self, nodes):
        """
        Get the bifurcation nodes in bewteen the provided input nodes

        Parameters
        ----------
        nodes: list of `neat.SNode`
            the input nodes

        Returns
        -------
        list of `neat.SNode`
            the bifurcation nodes
        """
        # unique nodes
        nodes = reduce(lambda l, x: l.append(x) or l if x not in l else l, nodes, [])
        # tag all nodes
        for node in self:
            node['tag'] = 0
        # find the 'leafs' within the list of nodes (i.e. most centripetal nodes)
        pnodes = []
        for node in nodes:
            pnodes.extend([n for n in self.pathToRoot(node)])
        pcount = Counter(pnodes)
        nodes = [node for node in nodes if pcount[node] == 1]
        # for each node, find nearest bifurcation towards root
        pnodes = []
        for node in nodes:
            pnodes.extend([n for n in self.pathToRoot(node)])
        for pathnode in pnodes:
            pathnode['tag'] += 1
        # find the bifurcation nodes
        bifur_nodes = [self.root]
        for node in self:
            # !!! only works for bifurcations with two children
            # TODO: extend for all bifurcations
            if len(node.child_nodes) > 1 and \
               not np.any([cn['tag'] == 0 for cn in node.child_nodes]):
                bifur_nodes.append(node)
        # remove the tags
        for node in self:
            del node.content['tag']
        # add root if necessary
        if self.root not in bifur_nodes:
            bifur_nodes = [self.root] + bifur_nodes

        return bifur_nodes

    def getNearestNeighbours(self, node, nodes):
        """
        Find the nearest neighbours of ``node`` in ``nodes``. If ``nodes`` contains
        ``node``, it is excluded from the search.

        When a node in the up-direction is a bifurcation node and in ``nodes``, nodes
        in its other subtree are excluded from the search

        !!! Untested

        Parameters
        ----------
        node: `neat.SNode`
            node for which the nearest neighbours are sought
        nodes: list of `neat.SNode`
            list in which nearest neighbours of ``node`` are sought
        """
        nns = []
        self._searchNNUp(node, nodes, nns)
        self._searchNNDown(node, nodes, nns)
        return nns

    def _searchNNUp(self, node, nodes, nns):
        p_node = node.parent_node
        if p_node is not None:
            # up direction
            if p_node in nodes:
                nns.append(p_node)
            else:
                self._searchNNUp(p_node, nodes, nns)
                # down direction
                for c_node in set(p_node.child_nodes) - {node}:
                    self._searchNNDown(c_node, nodes, nns)

    def _searchNNDown(self, node, nodes, nns):
        for c_node in node.child_nodes:
            if c_node in nodes:
                nns.append(c_node)
            else:
                self._searchNNDown(c_node, nodes, nns)

    def __copy__(self, new_tree=None):
        """
        Fill the ``new_tree`` with it's corresponding nodes in the same
        structure as ``self``, and copies all node variables that both tree
        classes have in common

        Parameters
        ----------
        new_tree: `neat.STree` or derived class (default is ``None``)
            the tree class in which the ``self`` is copied. If ``None``,
            returns a copy of ``self``.

        Returns
        -------
        The new tree instance
        """
        if new_tree is None:
            new_tree = self.__class__()

        new_node = new_tree._createCorrespondingNode(self.root.index)

        self.root.__copy__(new_node=new_node)
        new_tree.setRoot(new_node)
        self._recurseCopy(self.root, new_tree)
        # copy all attributes not related to tree structure
        orig_keys = set(self.__dict__.keys())
        copy_keys = orig_keys.intersection(set(new_tree.__dict__.keys()))
        for key in copy_keys:
            if key not in ['root', '_computational_root', '_original_root']:
                new_tree.__dict__[key] = copy.deepcopy(self.__dict__[key])

        return new_tree

    def _recurseCopy(self, pnode, new_tree):
        for node in pnode.getChildNodes(skip_inds=[]):
            new_node = new_tree._createCorrespondingNode(node.index)
            new_node = node.__copy__(new_node=new_node)
            new_tree.addNodeWithParent(new_node, new_tree.__getitem__(pnode.index,
                                                                      skip_inds=[]))
            self._recurseCopy(node, new_tree)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
# Adapted from mpl_toolkits.axes_grid2
# LICENSE: Python Software Foundation (http://docs.python.org/license.html)

from matplotlib.offsetbox import AnchoredOffsetbox
class AnchoredScaleBar(AnchoredOffsetbox):
    def __init__(self, transform, sizex=0, sizey=0, labelx=None, labely=None, loc=4,
                 pad=0.1, borderpad=0.1, sep=2, lw=2, prop=None, **kwargs):
        """
        Draw a horizontal and/or vertical  bar with the size in data coordinate
        of the give axes. A label will be drawn underneath (center-aligned).

        - transform : the coordinate frame (typically axes.transData)
        - sizex,sizey : width of x,y bar, in data units. 0 to omit
        - labelx,labely : labels for x,y bars; None to omit
        - loc : position in containing axes
        - pad, borderpad : padding, in fraction of the legend font size (or prop)
        - sep : separation between labels and bars in points.
        - **kwargs : additional arguments passed to base class constructor
        """
        from matplotlib.patches import Rectangle
        from matplotlib.offsetbox import AuxTransformBox, VPacker, HPacker, TextArea, DrawingArea
        bars = AuxTransformBox(transform)
        if sizex:
            bars.add_artist(Rectangle((0,0), sizex, 0, fc="none", lw=lw))
        if sizey:
            bars.add_artist(Rectangle((0,0), 0, sizey, fc="none", lw=lw))

        if sizex and labelx:
            bars = VPacker(children=[bars, TextArea(labelx, minimumdescent=False)],
                           align="center", pad=0, sep=sep)
        if sizey and labely:
            bars = HPacker(children=[TextArea(labely), bars],
                            align="center", pad=0, sep=sep)

        AnchoredOffsetbox.__init__(self, loc, pad=pad, borderpad=borderpad,
                                   child=bars, prop=prop, frameon=False, **kwargs)

def addScalebar(ax, matchx=True, matchy=True, hidex=True, hidey=True, **kwargs):
    """ Add scalebars to axes

    Adds a set of scale bars to *ax*, matching the size to the ticks of the plot
    and optionally hiding the x and y axes

    - ax : the axis to attach ticks to
    - matchx,matchy : if True, set size of scale bars to spacing between ticks
                    if False, size should be set using sizex and sizey params
    - hidex,hidey : if True, hide x-axis and y-axis of parent
    - **kwargs : additional arguments passed to AnchoredScaleBars

    Returns created scalebar object
    """
    def f(axis):
        l = axis.get_majorticklocs()
        return len(l)>1 and (l[1] - l[0])

    if matchx:
        kwargs['sizex'] = f(ax.xaxis)
        kwargs['labelx'] = r'' + str(kwargs['sizex']) + ' ' + kwargs['labelx']
    if matchy:
        kwargs['sizey'] = f(ax.yaxis)
        kwargs['labely'] = r'' + str(kwargs['sizey']) + ' ' + kwargs['labelx']

    sb = AnchoredScaleBar(ax.transData, **kwargs)
    ax.add_artist(sb)

    if hidex : ax.xaxis.set_visible(False)
    if hidey : ax.yaxis.set_visible(False)

    return sb



"""
Writes C++ simulation code for the ionchannels
"""

import os

# from neat.channels import channelcollection
from .channelcollection import channelcollection

def write():

    print('--> Writing c++ channel file')
    path = os.path.join(os.path.dirname(__file__), '../tools/simtools/net/')
    # path = '../tools/simtools/net/'

    fcc = open(os.path.join(path, 'Ionchannels.cc'), 'w')
    fh = open(os.path.join(path, 'Ionchannels.h'), 'w')
    fh.write('#include <iostream>' + '\n')
    fh.write('#include <string>' + '\n')
    fh.write('#include <vector>' + '\n')
    fh.write('#include <list>' + '\n')
    fh.write('#include <map>' + '\n')
    fh.write('#include <complex>' + '\n')
    fh.write('#include <string.h>' + '\n')
    fh.write('#include <stdlib.h>' + '\n')
    fh.write('#include <algorithm>' + '\n')
    fh.write('#include <math.h>' + '\n')
    fh.write('#include <time.h>' + '\n')
    fh.write('#include <time.h>' + '\n')
    fh.write('using namespace std;' + '\n\n')

    fh.write('class IonChannel{' + '\n')
    fh.write('protected:' + '\n')
    fh.write('    double m_g_bar = 0.0, m_e_rev = 50.00000000;' + '\n')
    fh.write('    bool m_instantaneous = false' + ';\n')
    fh.write('public:' + '\n')
    fh.write('    void init(double g_bar, double e_rev){m_g_bar = g_bar; m_e_rev = e_rev;};' + '\n')
    fh.write('    void setInstantaneous(bool b){m_instantaneous = b;};' + '\n')
    fh.write('    virtual void calcFunStatevar(double v){};' + '\n')
    fh.write('    virtual double calcPOpen(){return 0.0;};' + '\n')
    fh.write('    virtual void setPOpen(){};' + '\n')
    fh.write('    virtual void setPOpenEQ(double v){};' + '\n')
    fh.write('    virtual void advance(double dt){};' + '\n')
    fh.write('    virtual double getCond(){return 0.0;};' + '\n')
    fh.write('    virtual double getCondNewton(){return 0.0;};' + '\n')
    fh.write('    virtual double f(double v){return 0.0;};' + '\n')
    fh.write('    virtual double DfDv(double v){return 0.0;};' + '\n')
    fh.write('    virtual void setfNewtonConstant(double* vs, int v_size){};' + '\n')
    fh.write('    virtual double fNewton(double v){return 0.0;};' + '\n')
    fh.write('    virtual double DfDvNewton(double v){return 0.0;};' + '\n')
    fh.write('};' + '\n')


    fcc.write('#include "Ionchannels.h"' + '\n')
    fh.write('\n')
    fcc.write('\n')
    fcc.close()
    fh.close()

    for name, channel_class in list(channelcollection.__dict__.items()):
        if isinstance(channel_class, type) and \
           (name != 'IonChannel' and name != 'IonChannelSimplified') and '_func' not in name:
            chan = channel_class()
            chan.writeCPPCode(path)


    fh = open(os.path.join(path, 'Ionchannels.h'), 'a')
    fcc = open(os.path.join(path, 'Ionchannels.cc'), 'a')
    fh.write('class ChannelCreator{\n')
    fh.write('public:\n')
    fh.write('    IonChannel* createInstance(string channel_name){\n')
    kk = 0
    for name, channel_class in list(channelcollection.__dict__.items()):
        if isinstance(channel_class, type) and \
           (name != 'IonChannel' and name != 'IonChannelSimplified') and '_func' not in name:
            if kk == 0:
                fh.write('        if(channel_name == \"' + name +'\"){\n')
            else:
                fh.write('        else if(channel_name == \"' + name +'\"){\n')
            fh.write('            return new ' + name + '();\n')
            fh.write('        }\n')
            kk += 1
    fh.write('    };' + '\n')
    fh.write('};' + '\n')

    fh.close()
    fcc.close()



"""
Module for finding kernels and fitting them with sums of exponentials.

Uses the Vector Fitting algorithm (Gustavsen, 1999) and Prony method.

Author: Willem Wybo
Date: 13/05/2016
"""


import numpy as np
import scipy.linalg as la
from scipy.cluster.vq import kmeans

import math
import copy



class Fitter(object):
    def der(self, x, arr):
        dx = x[1] - x[0]
        diffarr = (arr[1:] - arr[0:-1]) / dx
        return diffarr, x[0:-1] + dx/2

    def derder(self, x, arr):
        dx = x[1] - x[0]
        diffarr, _ = self.der(x, arr)
        diffdiffarr = (diffarr[1:] - diffarr[0:-1]) / dx
        return diffdiffarr, x[1:-1]

    def zerocrossing(self, x, arr):
        arr = copy.copy(arr)
        inds = np.where(np.diff(np.sign(arr)))[0]
        return inds, x[inds]

    def find_nearest(self, array, value):
        idx = (np.abs(array-value)).argmin()
        return array[idx], idx


class ExpFitter(Fitter):
    def sumExp(self, x, a, c, flat=True):
        if flat:
            return (np.exp(x[:,None]*a[:,None].T).dot(c[:,None])).flatten().real
        else:
            return np.exp(x[:,None]*a[:,None].T).dot(c[:,None]).real

    def PronyExpFit(self, deg, x, y):
        """
        Construct a sum of exponentials fit to a given time-sequence y by
        using prony's method

        input:
            [deg]: int, number of exponentials
            [x]: numpy array, sequence of regularly spaced points at which y is evaluated
            [y]: numpy array, sequence

        output:
            [a]: numpy array, exponential coefficient
            [c]: numpy array, exponentail magnitudes
            [rms]: float, root mean square error of the data
        """
        # deg += 1
        # stepsize
        h = x[1] - x[0]
        #Build matrix
        A = la.hankel(y[:-deg],y[-deg-1:])
        a = -A[:,:deg]
        b = A[:,deg]
        #Solve it
        s = la.lstsq(a,b)[0]
        #Solve polynomial
        p = np.flipud(np.hstack((s,1)))
        u = np.roots(p)
        #Only keep roots in unit circle
        inds = np.where(np.logical_and((np.abs(u) < 1.), \
                np.logical_not(np.logical_and(np.imag(u) == 0., np.real(u) <= 0.))))[0]
        u = u[inds]
        #Calc exponential factors
        a = np.log(u)/h
        #Build power matrix
        A = np.power((np.ones((len(y),1))*u[:,None].T),np.arange(len(y))[:,None]*np.ones((1,len(inds))))
        #solve it
        f = la.lstsq(A,y)[0]
        #calc amplitudes
        c = f/np.exp(a*x[0])
        #build x, approx and calc rms
        approx = self.sumExp(x, a, c).real
        rms = np.sqrt(((approx-y)**2).sum() / len(y))
        return a, c, rms

    def construct_Hankel_matrices(self, y):
        ind0 = int(len(y)/2)
        # original and shifted hankel matrix
        H0 = la.hankel(y[0:ind0], y[ind0-1:2*ind0-1])
        H1 = la.hankel(y[1:ind0+1], y[ind0:2*ind0])
        return H0, H1

    # def Z_McE_ExpFit(self, x, y, deg=2):
    #     # construct the Hankel matrices
    #     H0, H1 = self.construct_Hankel_matrices(y)
    #     # compute the singular value decomposition
    #     U, s, Vh = la.svd(H0)
    #     U_ = U[:, 0:deg]
    #     Vh_ = Vh[0:deg, :]
    #     s_ = s[0:deg]
    #     # compute system matrix
    #     F0 = np.diag(1./np.sqrt(s_)).dot(U_.T)
    #     F1 = Vh_.T.dot(np.diag(1./np.sqrt(s_)))
    #     A = F0.dot(H1.dot(F1))
    #     # find eigenvalues of system matrix
    #     u, v = la.eig(A)
    #     # system time-scales (inverse)
    #     alphas = np.log(u) / dx
    #     return alphas

    def fitExp_Z_McE(self, x, y, rtol=1e-2, maxdeg=10):
        deg = 1; rms = 1.
        # stepsize
        dx = x[1] - x[0]
        # construct the Hankel matrices
        H0, H1 = self.construct_Hankel_matrices(y)
        # compute the singular value decomposition
        U, s, Vh = la.svd(H0)
        # loop over increasing number of exponentials
        while rms > rtol and deg < maxdeg:
            U_ = U[:, 0:deg]
            Vh_ = Vh[0:deg, :]
            s_ = s[0:deg]
            # compute system matrix
            F0 = np.diag(1./np.sqrt(s_)).dot(U_.T)
            F1 = Vh_.T.dot(np.diag(1./np.sqrt(s_)))
            A = F0.dot(H1.dot(F1))
            # find eigenvalues of system matrix
            u, v = la.eig(A)
            # system time-scales (inverse)
            alphas = np.log(u.real) / dx
            # solve weights
            A = np.exp(x[:,None] * alphas[None, :] * dx)
            gammas = la.lstsq(A,y)[0]
            # compute rmse
            approx = self.sumExp(x, alphas, gammas)
            rms = np.sqrt(((approx-y)**2).sum() / len(y))
            # increase degree
            deg += 1

        return alphas, gammas, rms

    def reduceSeries(self, a, c, x, y, rtol=1e-2):
        """
        Reduces the number of exponential terms in a series, till a given tolerance
        is reached

        input:
            [a]: numpy array of exponential timescales
            [c]: numpy array of exponential magnitudes
            [x]: numpy array of x-values at which the function is evaluated
            [y]: numpy array of function values
            [rtol]: float, relative tolerance given the largest function value
        output:
            [alpha]: exponential coefficients
            [gamma]: magnitudes
            [rms]: float, root mean square error
        """
        k = 1; rms = 2*rtol

        while rms > rtol and k <= len(a):
            sortind = np.argsort(np.abs(c))[::-1]
            alpha = a[sortind][0:k]
            gamma = c[sortind][0:k]

            approx = self.sumExp(x, alpha, gamma).real
            rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
            k += 1

        return alpha, gamma, rms

    def fitExp(self, x, y, deg=30, rtol=1e-2, surface=False, A=None):
        a, c, rms = self.PronyExpFit(deg, x, y)
        alpha, gamma, rms = self.reduceSeries(a, c, x, y, rtol=rtol)
        if surface:
            dx = x[1] - x[0]
            if A == None:
                A = dx * np.sum(y)
            Afit = np.sum(gamma * (np.exp(alpha*x[-1]) - np.exp(alpha*x[0])) / alpha)
            gamma = gamma * A / Afit
            approx = self.sumExp(x, alpha, gamma).real
            rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))

        return alpha, gamma, rms


class fExpFitter(Fitter):
    def sumFExp(self, s, alphas, gammas):
        return np.sum(self.fexps(s, alphas, gammas), 0)

    def fexps(self, s, alphas, gammas):
        return gammas[:,None] / (alphas[:,None] + s[None,:])

    def trialFunFit(self, s, arr, alphas, pairs=None):
        # construct matrix for extended fitting problem
        A = np.concatenate((1. / (s[:,None] + alphas[None,:]), \
                                arr[:,None] / (s[:,None] + alphas[None,:])), axis=1)
        # find auxiliary residues
        c = la.lstsq(A, arr)[0][-len(alphas):]
        # find zeros of fitted auxiliary function
        H = np.diag(alphas) - np.dot(np.ones((len(alphas),1), dtype=complex), c[None,:])
        alphanew = np.linalg.eig(H)[0]
        # find real residues
        Anew = 1. / (s[:,None] + alphanew[None,:])
        cnew = la.lstsq(Anew, arr)[0]

        return alphanew, cnew, None

    def trialFunFit_constrained(self, s, arr, alphas, pairs, zerostart=False):
        deg = len(alphas)
        carr = np.concatenate((arr.real, arr.imag))
        # construct matrix for extended fitting problem
        A = np.concatenate((1. / (s[:,None] + alphas[None,:]), \
                                arr[:,None] / (s[:,None] + alphas[None,:])), axis=1)
        # implement the constraint
        pairsnew = np.concatenate((pairs, pairs))
        for i, p in enumerate(pairsnew):
            if p:
                x1 = A[:,i] + A[:,i+1]
                x2 = 1j * (A[:,i] - A[:,i+1])
                A[:,i] = x1
                A[:,i+1] = x2
        A = np.concatenate((A.real, A.imag), axis=0)
        # find auxiliary residues
        c = la.lstsq(A, carr)[0][-len(alphas):]
        # find zeros of fitted auxiliary function
        a = np.diag(alphas)
        b = np.ones(deg)
        # implement similarity transform
        for i, p in enumerate(pairs):
            if p:
                a[i:i+2, i:i+2] = np.array([[alphas[i].real, alphas[i].imag], \
                                            [-alphas[i].imag, alphas[i].real]])
                b[i:i+2] = np.array([2,0])
        H = a.real - np.dot(b[:,None], c[None,:])
        alphanew = np.linalg.eig(H)[0]
        inds = np.argsort(alphanew)
        alphanew = alphanew[inds]
        # indicates where pairs of complex conjugate poles occur
        auxarr = np.abs((np.abs(alphanew[:-1]) - np.abs(alphanew[1:])) / np.abs(alphanew[:-1]))
        auxarr2 = np.abs(alphas.imag) > 1e-15
        pairs = np.logical_and(np.concatenate((auxarr < 1e-15, np.zeros(1, dtype=bool))), auxarr2)
        # find residues
        Anew = 1. / (s[:,None] + alphanew[None,:])
        for i, p in enumerate(pairs):
            if p:
                x1 = Anew[:,i] + Anew[:,i+1]
                x2 = 1j * (Anew[:,i] - Anew[:,i+1])
                Anew[:,i] = x1
                Anew[:,i+1] = x2
        Anew = np.concatenate((Anew.real, Anew.imag), axis=0)
        if zerostart:
            # enforce K(t=0)=0 constraint
            row1 = np.ones(2*deg)
            for i, p in enumerate(pairs):
                if p:
                    row1[i+1] = 0
            Anew = np.concatenate((np.ones((1, deg), dtype=complex), Anew), axis=0)
            carr = np.concatenate((np.zeros(1, dtype=complex), carr))
        cnew = la.lstsq(Anew, carr)[0]
        cnew = np.array(cnew, dtype=complex)
        # recast cnew to complex values
        for i, p in enumerate(pairs):
            if p:
                cnew[i:i+2] = np.array([cnew[i] + 1j * cnew[i+1], cnew[i] - 1j * cnew[i+1]])

        return alphanew, cnew, pairs

    def fit_residues(self, s, arr, alphas, pairs):
        carr = np.concatenate((arr.real, arr.imag))
        A = 1. / (s[:,None] + alphas[None,:])
        for i, p in enumerate(pairs):
            if p:
                x1 = A[:,i] + A[:,i+1]
                x2 = 1j * (A[:,i] - A[:,i+1])
                A[:,i] = x1
                A[:,i+1] = x2
        A = np.concatenate((A.real, A.imag), axis=0)
        cnew = la.lstsq(A, carr)[0]
        cnew = np.array(cnew, dtype=complex)
        # recast cnew to complex values
        for i, p in enumerate(pairs):
            if p:
                cnew[i:i+2] = np.array([cnew[i] + 1j * cnew[i+1], cnew[i] - 1j * cnew[i+1]])
        return cnew

    def trialFunFit_constrained_2d(self, s, arr2d, alphas, pairs):
        print('>>> multifun fit test v2 <<<')
        deg = len(alphas)
        # construct f array
        arr1d = np.array([], dtype=complex)
        for ind, arr in enumerate(arr2d):
            arr1d = np.concatenate((arr1d, arr))
        # construct matrix A
        ns = len(s)
        ncols = (len(arr2d) + 1) * deg
        nrows = len(arr1d)
        A = np.zeros((nrows, ncols), dtype=complex)
        for ind, fis in enumerate(arr1d):
            indA = int(ind/ns)
            A[ind,deg*indA:deg*(indA+1)] = 1./(s[ind%ns] + alphas)
            # try:
            #     A[ind,deg*indA:deg*(indA+1)] = 1./(s[ind%ns] + alphas)
            # except ValueError:
            #     print indA
            #     print deg*indA
            #     print deg*(indA+1)
            #     print ncols
            A[ind,-deg:] = -fis / (s[ind%ns] + alphas)
        # implement the constraint
        for j in range(len(arr2d) + 1):
            for i, p in enumerate(pairs):
                if p:
                    x1 = A[:,j*deg+i] + A[:,j*deg+i+1]
                    x2 = 1j * (A[:,j*deg+i] - A[:,j*deg+i+1])
                    A[:,j*deg+i] = x1
                    A[:,j*deg+i+1] = x2
        A = np.concatenate((A.real, A.imag), axis=0)
        arr1d = np.concatenate((arr1d.real, arr1d.imag))
        # find auxiliary residues
        c = la.lstsq(A, arr1d)[0][-len(alphas):]
        print('cnew: ', c)
        # find zeros of fitted auxiliary function
        a = np.diag(alphas)
        b = np.ones(deg)
        # implement similarity transform
        for i, p in enumerate(pairs):
            if p:
                a[i:i+2, i:i+2] = np.array([[alphas[i].real, alphas[i].imag], \
                                            [-alphas[i].imag, alphas[i].real]])
                b[i:i+2] = np.array([2,0])
        # compute zeros of sum sigmafit
        H = a.real - np.dot(b[:,None], c[None,:])
        print('H: ', H)
        alphanew = np.linalg.eig(H)[0]
        print('alphanew: ', alphanew)
        inds = np.argsort(alphanew)
        alphanew = alphanew[inds]
        # indicates where pairs of complex conjugate poles occur
        auxarr = np.abs((np.abs(alphanew[:-1]) - np.abs(alphanew[1:])) / np.abs(alphanew[:-1]))
        auxarr2 = np.abs(alphanew.imag) > 1e-15 # np.abs(alphas.imag) > 1e-15
        pairs = np.logical_and(np.concatenate((auxarr < 1e-15, np.zeros(1, dtype=bool))), auxarr2)
        # find residues
        # compute matrix for residue calculation
        Anew = 1. / (s[:,None] + alphanew[None,:])
        for i, p in enumerate(pairs):
            if p:
                x1 = Anew[:,i] + Anew[:,i+1]
                x2 = 1j * (Anew[:,i] - Anew[:,i+1])
                Anew[:,i] = x1
                Anew[:,i+1] = x2
        Anew = np.concatenate((Anew.real, Anew.imag), axis=0)
        # compute residues
        c2dnew = np.zeros((arr2d.shape[0], deg), dtype=complex)
        for ind, arr in enumerate(arr2d):
            carr = np.concatenate((arr.real, arr.imag))
            cnew = la.lstsq(Anew, carr)[0]
            cnew = np.array(cnew, dtype=complex)
            # recast cnew to complex values
            for i, p in enumerate(pairs):
                if p:
                    cnew[i:i+2] = np.array([cnew[i] + 1j * cnew[i+1], cnew[i] - 1j * cnew[i+1]])
            c2dnew[ind,:] = cnew

        print('cnew: ', c2dnew)

        return alphanew, c2dnew, pairs

    def reduceSeries(self, s, y, a, c, pairs=None, rtol=1e-2, pprint=True):
        """
        reduce the series of exponentials after the fitting
        """
        k = 1; rms = 1.

        # ensure stability of approximation
        inds = np.where(a.real > 0.)[0]
        a = a[inds]; c = c[inds];
        if pairs is not None: pairs = pairs[inds]

        # construct indices for ranking the exponentials
        pairs_alltrue = copy.copy(pairs)
        for i,p in enumerate(pairs):
            if p: pairs_alltrue[i+1] = True
        magnitudes = np.zeros(a.shape)
        for i in range(len(pairs_alltrue)):
            if pairs_alltrue[i]:
                c_ = c[i].real; c__ = c[i].real
                a_ = a[i].real; a__ = a[i].real
                magnitudes[i] = (c_*a_ + c__*a__) / (a_**2 + a__**2)
            else:
                magnitudes[i] = c[i].real / a[i].real

        sortind = np.argsort(np.abs(magnitudes))[::-1]

        anew = copy.copy(a[sortind])
        alphas = anew
        cnew = copy.copy(c[sortind])
        gammas = cnew
        # look for pairs to be sure they are correct
        auxarr = np.abs((np.abs(alphas[:-1]) - np.abs(alphas[1:])) / np.abs(alphas[:-1]))
        auxarr2 = np.abs(alphas.imag) > 1e-15
        pairs = np.logical_and(np.concatenate((auxarr < 1e-15, np.zeros(1, dtype=bool))), auxarr2)
        npairs = copy.copy(pairs)

        approx = self.sumFExp(s, alphas, gammas)

        while rms > rtol and k < len(a)+1:
            if (pairs is not None) and pairs[k-1]:
                k += 1
            alphas = anew[0:k]
            gammas = cnew[0:k]

            auxarr = np.abs((np.abs(alphas[:-1]) - np.abs(alphas[1:])) / np.abs(alphas[:-1]))
            auxarr2 = np.abs(alphas.imag) > 1e-15
            npairs = np.logical_and(np.concatenate((auxarr < 1e-15, np.zeros(1, dtype=bool))), auxarr2)

            approx = self.sumFExp(s, alphas, gammas)
            rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
            k += 1

        if pprint:
            pairinds = copy.copy(npairs)
            inds = np.where(npairs)[0]
            for i in inds:
                pairinds[i+1] = True
            inds = np.where(np.logical_not(pairinds))[0]
            if len(inds) > 0:
                if np.max(np.abs(alphas[inds].imag)) > 1e-6:
                    print('!!! Warning: invalid pairs !!!')
                    print('original alphas: ', anew)
                    print('original gammas: ', cnew)
                    print('original pairs: ', pairs)
                    print('new alphas: ', alphas)
                    print('new gammas: ', gammas)
                    print('new pairs: ', npairs)

        return alphas, gammas, rms, approx, npairs

    def _find_start_nodes(self, s, deg, realpoles, initpoles):
        if initpoles == 'lin':
            trialpoles = np.linspace(s[int(len(s)/2.)+1].imag, s[-1].imag, deg)
        elif initpoles == 'log10':
            trialpoles = np.logspace(1, np.log10(s[-1].imag), num=deg, base=10)
        elif initpoles == 'log':
            trialpoles = np.logspace(1, np.log(s[-1].imag), num=deg, base=math.e)
        elif initpoles == 'random':
            trialpoles = s[-1].imag * np.random.rand(deg)
        else:
            raise Exception('initpoles invalid')
        if realpoles:
            pairs = np.zeros(trialpoles.shape, dtype=bool)
        else:
            trialpoles = np.array([[tp + 1j*tp, tp - 1j*tp] for tp in trialpoles]).flatten()
            pairs = np.array([[True, False] for _ in range(deg)]).flatten()
        return trialpoles, pairs

    def _run_fit(self, s, y, trialpoles, pairs, rtol, maxiter, constrained, zerostart, pole_flip=True, pprint=True):
        """
        performs iterations of the actual fitting process
        """
        k = 0; rms = rtol+1.
        l = 0; m = 0

        alist = []; clist = []; rmslist = []; pairslist = []

        trialpoles_orig = copy.copy(trialpoles)
        pairs_orig = copy.copy(pairs)

        if constrained:
            while rms > rtol and k < maxiter:
                a, c, pairs = self.trialFunFit_constrained(s, y, trialpoles, pairs, zerostart=zerostart)
                approx = self.sumFExp(s, a, c)
                rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
                # if unstable poles, make sure to run again
                # if np.min(a) < 0.:
                #     rms = rtol + 1.
                #     if m < 10.:
                #         if pole_flip:
                ind = np.where(a < 0.)[0] # where poles are unstable
                if len(ind) > 0:
                    a[ind] *= -1.
                    c = self.fit_residues(s, y, a, pairs)
                    approx = self.sumFExp(s, a, c)
                    rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
                            # if rms < rtol:
                            #     alist.append(copy.deepcopy(a)); clist.append(copy.deepcopy(c)); rmslist.append(rms); pairslist.append(pairs)
                        # else:
                        #     ind = np.where(a > 0.)[0] # where poles are stable
                        #     newpole, newpair = self._find_start_nodes(s, len(a)-len(ind), True, 'random')
                        #     trialpoles = np.concatenate((a[ind], newpole))
                        #     pairs = np.concatenate((pairs[ind], newpair))
                    # else:
                    #     trialpoles, pairs = self._find_start_nodes(s, len(trialpoles_orig), True, 'random')
                    #     m = 0
                    # l += 1; m += 1
                # else:
                alist.append(copy.deepcopy(a)); clist.append(copy.deepcopy(c)); rmslist.append(rms); pairslist.append(pairs)
                trialpoles = copy.copy(a)
                k += 1
            if pprint and l > 5:
                print('Often found unstable poles (' + str(l) + ' times)')
            return alist, clist, rmslist, pairslist
        else:
            while rms > rtol and k < maxiter:
                a, c,_ = self.trialFunFit(s, y, trialpoles, zerostart)
                approx = self.sumFExp(s, a, c)
                rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
                trialpoles = a
                k += 1
            return alist, clist, rmslist, None

    def _run_fit_vector(self, s, ys, trialpoles, pairs, rtol, maxiter):
        # eps = 2.
        # k = 0; rms = 1.
        # rms_ = rms

        # alist = []; clist = []; rmslist = []; pairslist = []

        # while rms > rtol and k < maxiter:

        #     a, c2d, pairs = self.trialFunFit_constrained_2d(s, ys, trialpoles, pairs)
        #     rms = 0.
        #     for ind, y in enumerate(ys):
        #         approx = self.sumFExp(s, a, c2d[ind])
        #         rms += np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
        #     alist.append(copy.deepcopy(a)); clist.append(copy.deepcopy(c2d)); rmslist.append(rms); pairslist.append(pairs)
        #     # randomize poles a bit
        #     skip = False
        #     tp = copy.deepcopy(a)
        #     if (rms_ - rms) / rms_ < eps:
        #         for i, p in enumerate(pairs):
        #             if not skip:
        #                 if p:
        #                     x1 = 0.1 * tp[i].real; x2 = 0.1 * np.abs(tp[i].imag)
        #                     r1 = x1 * (2. * np.random.rand() - 1); r2 = x2 * (2. * np.random.rand() - 1)
        #                     tp[i:i+2] = np.array([tp[i] + r1 + 1j*r2, tp[i+1] + r1 - 1j*r2])
        #                     skip = True
        #                 else:
        #                     x = 0.1 * tp[i]
        #                     r = x * (2. * np.random.rand() - 1)
        #                     tp[i] += r
        #                     skip = False
        #     trialpoles = tp
        #     k += 1
        #     rms_ = rms
        eps = 2.
        k = 0; rms = 1.
        rms_ = rms

        alist = []; clist = []; rmslist = []; pairslist = []

        while rms > rtol and k < maxiter:
            a2d = np.zeros((len(ys), len(trialpoles)), dtype=complex)
            c2d = np.zeros((len(ys), len(trialpoles)), dtype=complex)
            pairs2d = np.zeros((len(ys), len(trialpoles)), dtype=bool)
            for ind, y in enumerate(ys):
                a2d[ind], c2d[ind], pairs2d[ind] = self.trialFunFit_constrained(s, y, trialpoles, pairs)
                # put complex conjugates with positive part first
                for i, p in enumerate(pairs2d[ind]):
                    if p:
                        if a2d[ind,i] < 0:
                            a2d[ind,i] = a2d[ind,i].real - 1j * a2d[ind,i].imag
                            a2d[ind,i+1] = a2d[ind,i+1].real - 1j * a2d[ind,i+1].imag
                            c2d[ind,i] = c2d[ind,i].real - 1j * c2d[ind,i].imag
                            c2d[ind,i+1] = c2d[ind,i+1].real - 1j * c2d[ind,i+1].imag
            a, pairs = self._Kmeans(a2d, pairs2d)
            c2d = np.zeros((len(ys), len(a)), dtype=complex)
            for ind, y in enumerate(ys):
                c2d[ind] = self.fit_residues(s, y, a, pairs)
                approx = self.sumFExp(s, a, c2d[ind])
                rms += np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
            alist.append(copy.deepcopy(a)); clist.append(copy.deepcopy(c2d)); rmslist.append(rms); pairslist.append(pairs)
            # randomize poles a bit
            skip = False
            tp = copy.deepcopy(a)
            if (rms_ - rms) / rms_ < eps:
                for i, p in enumerate(pairs):
                    if not skip:
                        if p:
                            x1 = 0.1 * tp[i].real; x2 = 0.1 * np.abs(tp[i].imag)
                            r1 = x1 * (2. * np.random.rand() - 1); r2 = x2 * (2. * np.random.rand() - 1)
                            tp[i:i+2] = np.array([tp[i] + r1 + 1j*r2, tp[i+1] + r1 - 1j*r2])
                            skip = True
                        else:
                            x = 0.1 * tp[i]
                            r = x * (2. * np.random.rand() - 1)
                            tp[i] += r
                            skip = False
            trialpoles = tp
            k += 1
        return alist, clist, rmslist, pairslist

    def _Kmeans(self, a2d, pairs2d):        # do the kmeans algorithm to make sure all nodes are the same
        a1d = np.array([], dtype=complex)
        for i, a in enumerate(a2d):
            # determine the coefficients not to take into account in the algorithm
            paux = np.concatenate((np.array([False]), pairs2d[i,:-1]))
            inds = np.where(np.logical_not(paux))[0]
            a1d = np.concatenate((a1d,a[inds]))
        adata = np.concatenate((a1d.real[:,None], a1d.imag[:,None]), 1)
        astart = np.concatenate((a2d[-1].real[inds][:,None], a2d[-1].imag[inds][:,None]), 1)
        a = kmeans(adata, astart)[0]
        # check for complex conjugates
        anew = []; pairsnew = []
        for alpha in a:
            if np.abs(alpha[1]) > 1e-9:
                anew.append(alpha[0] + 1j*alpha[1])
                anew.append(alpha[0] - 1j*alpha[1])
                pairsnew.append(True)
                pairsnew.append(False)
            else:
                anew.append(alpha[0] + 1j*0.)
                pairsnew.append(False)
        # a = a[:,0] + 1j* a[:,1]
        # look for pairs to be sure they are correct

        # auxarr = np.abs((np.abs(a[:-1]) - np.abs(a[1:])) / np.abs(a[:-1]))
        # auxarr2 = np.abs(a.imag) > 1e-15
        # pairs = np.logical_and(np.concatenate((auxarr < 1e-15, np.zeros(1, dtype=bool))), auxarr2)
        return np.array(anew), np.array(pairsnew)

    def reduceNumExp(self, s, y, a, c, pairs, lim=0.1, pprint=True, pplot=True):
        """
        pools the short timescale exponentials
        """
        # find inds of exponentials that have to be taken together
        inds = np.where(np.abs(a.real) > (1e3 / lim))[0]
        # the other indices stay the same
        inds_no = np.where(np.abs(a.real) <= (1e3 / lim))[0]
        anew = a[inds_no]; cnew = c[inds_no]; pairsnew = pairs[inds_no]
        if len(inds) > 1:
            amin = np.min(a[inds])
            EF = ExpFitter()
            if pplot == True:
                import matplotlib.pyplot as pl
                y_f_full = self.sumFExp(s, a, c)
                y_f_part = self.sumFExp(s, a[inds], c[inds])
                pl.figure('reduceNumExp problem')
                pl.plot(s.imag, y_f_full.real, 'r')
                pl.plot(s.imag, y_f_part.real, 'b')
                pl.plot(s.imag, y_f_full.imag, 'r--')
                pl.plot(s.imag, y_f_part.imag, 'b--')
                pl.show()
            # multiple step approach
            t = np.linspace(0., 5./amin.real, 1000.)
            y_t = EF.sumExp(t, -a[inds], c[inds])
            y_t_full = EF.sumExp(t, -a, c)
            A_t = -np.sum(c[inds] * (np.exp(-a[inds]*t[-1]) - np.exp(-a[inds]*t[0])) / a[inds])
            y_t_lim = EF.sumExp(np.array([lim*1e-3]), -a[inds], c[inds])
            y_t_lim_full = EF.sumExp(np.array([lim*1e-3]), -a, c)
            #~ print 'full sum at 1ms: ', y_t_lim_full
            #~ print 'partial sum at 1ms: ', y_t_lim
            #~ print 'max full sum: ', np.max(y_t_full[1:])
            # fit first outside of first timestep if necessary
            if amin.real < (2e4 / lim) and np.abs(y_t_lim_full - y_t_lim) > 0.001 * np.max(y_t_full[1:]):
                t_out = np.linspace(lim*1e-3, 5./amin.real, 1000.)
                y_out = EF.sumExp(t_out, -a[inds], c[inds])
                A_out = -np.sum(c[inds] * (np.exp(-a[inds]*t_out[-1]) - np.exp(-a[inds]*t_out[0])) / a[inds])
                try:
                    # if the maximum of the to be grouped exponentials is past lim,
                    # we use two exponentials, otherwise one
                    #~ else:
                    A, C, _ = EF.fitExp(t_out, y_out, deg=1, rtol=0.0001, surface=True, A=A_out)
                    A = -A.real; C = C.real
                    Ptemp = [False]
                except ValueError:
                    A = np.array([amin.real])
                    C = A_out / ((np.exp(-A*t_out[-1]) - np.exp(-A*t_out[0])) / A)
                    Ptemp = [False]
                # check if we need to fit inside first timestep
                t_in = np.linspace(0., lim*1e-3, 100)
                y_in = EF.sumExp(t_in, -a[inds], c[inds]) - EF.sumExp(t_in, -A, C)
                A_in_full = -np.sum(c[inds] * (np.exp(-a[inds]*t_in[-1]) - np.exp(-a[inds]*t_in[0])) / a[inds])
                A_in = -np.sum(C * (np.exp(-A*t_in[-1]) - np.exp(-A*t_in[0])) / A)
                if np.abs(A_in - A_in_full) < 0.01 * np.abs(A_in_full):
                    # we don't need to fit an extra exponential,
                    # but just rescale surfaces a bit
                    A_tot = np.sum(c[inds] / a[inds])
                    A_part = np.sum(C / A)
                    C = C * A_tot / A_part
                    P = np.array(Ptemp, dtype=bool)
                else:
                    # we need to fit an extra exponential
                    t = np.linspace(0., 3./amin.real, 1000.)

                    A_t = np.sum(c[inds] / a[inds])
                    A_exp1 = np.sum(C / A)
                    A2 =  np.array([1e4 / lim], dtype=complex)
                    C2 =  (A_t-A_exp1) * A2

                    P = np.array(Ptemp + [False], dtype=bool)
                    A = np.concatenate((A, A2))
                    C = np.concatenate((C, C2))
            else:
                # we can just fit inside the first timestep
                # construct new exponential naively
                A = np.array([amin.real], dtype=complex)
                C = np.sum(c[inds] / a[inds]) * A
                P = np.array([False], dtype=bool)

            # concatenate the arrays
            anew = np.concatenate((anew, A))
            cnew = np.concatenate((cnew, C))
            pairsnew = np.concatenate((pairsnew, P))

            if pprint or pplot:
                t = np.linspace(0.,0.050,100000)
                A_original = - np.sum(c * (np.exp(-a*t[-1]) - np.exp(-a*t[0])) / a)
                A_new = - np.sum(cnew * (np.exp(-anew*t[-1]) - np.exp(-anew*t[0])) / anew)
                if np.abs(A_original - A_new) > 1e-12 or np.isnan(A_new.real):
                    print('!!! Warning: surfaces under kernels not equal !!!')
                    print('oringal surface: ', A_original)
                    print('new surface: ', A_new)
                    print('all a\'s: ', a)
                    print('all gamma\'s: ', c)
                    print('all pairs: ', pairs)
                    print('tbg a\'s: ', a[inds])
                    print('tbg gamma\'s: ', c[inds])
                    print('tbg pairs: ', pairs[inds])
                    print('ntbg a\'s: ', a[inds_no])
                    print('ntbg gamma\'s: ', c[inds_no])
                    print('ntbg pairs: ', pairs[inds_no])
                    print('new a\'s: ', anew)
                    print('new c\'s: ', cnew)
                    print('new pairss: ', pairsnew)

            if pplot and (np.abs(A_original - A_new) > 1e-12 or np.isnan(A_new.real)):
            #~ if pplot:
                t = np.linspace(0.,0.050,100000)
                dt = t[1] - t[0]
                ef = ExpFitter()
                se_ = ef.sumExp(t, -a[inds], c[inds])
                e_ = ef.sumExp(t, -A, C)
                se = ef.sumExp(t, -a, c)
                e = ef.sumExp(t, -anew, cnew)
                print('integral original reduced: ', - np.sum(c[inds] * (np.exp(-a[inds]*t[-1]) - np.exp(-a[inds]*t[0])) / a[inds]))
                print('integral fit reduced: ', - np.sum(C * (np.exp(-A*t[-1]) - np.exp(-A*t[0])) / A))
                print('final a\'s :', anew)
                print('new a\'s :', A)
                import matplotlib.pyplot as pl
                pl.figure('reduce_exp problem')
                pl.plot(t*1000, se, 'r', label='original kernel')
                pl.plot(t*1000, e, 'b', label='new kernel')
                pl.plot(t*1000, se_, 'r--', label='exps to be reduced')
                pl.plot(t*1000, e_, 'b--', label='to be reduced exp')
                pl.legend(loc=0)
                pl.show()

        # new approximation and rmse
        approx = self.sumFExp(s, anew, cnew)
        rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))

        return anew, cnew, rms, approx, pairsnew

    def fitFExp_increment(self, s, y, rtol=1e-2, maxiter=20, maxiter_step=3, realpoles=True, constrained=True, zerostart=False, pprint=True):

        # find the start nodes
        trialpoles, pairs = self._find_start_nodes(s, 1, realpoles, 'log10')

        for k in range(maxiter):
            alist, clist, rmslist, pairslist \
                    = self._run_fit(s, copy.copy(y), trialpoles, pairs, rtol, maxiter_step,
                                    constrained, zerostart, pole_flip=True, pprint=pprint)
            indmin = np.argmin(np.array(rmslist))
            alpha = alist[indmin]; gamma = clist[indmin]; rms = rmslist[indmin]; pairs = pairslist[indmin]
            if rms < rtol:
                break
            else:
                if realpoles:
                    # alphanew = [s[-1].imag * np.random.rand()]
                    alphanew = [np.random.choice(1./s.imag)]
                    pairsnew = [False]
                else:
                    areal = s[-1].imag * np.random.rand()
                    aimag = s[-1].imag * np.random.rand()
                    alphanew = [areal + 1j*aimag, areal - 1j*aimag]
                    pairsnew = [True, False]
            trialpoles = np.array(alpha.tolist() + alphanew)
            # print trialpoles
            pairs = np.array(pairs.tolist() + pairsnew)
        rmsfinal = rms

        if pprint and rmsfinal > rtol:
            print('Target accuracy was not reached')

        return alpha, gamma, pairs, rmsfinal


    def fitFExp(self, s, y, deg=20, rtol=1e-2, maxiter=5, lim=None, realpoles=True, initpoles='lin',
                zerostart=False, constrained=True, reduce_numexp=False, return_real=False, lower_limit=None):
        """
        Fits a function in fourrierspace by a series of fourrier transformed exponentials.

        input:
            -args
            [s]: numpy array of frequencies (imaginary) at which value function is evaluated
            [y]: numpy array of complex function values
            -kwargs
            [deg]: int, number of exponential terms used (real number is dubbeled if realpoles=False)
            [rtol]: float, relative toleranse after which iterations stop
            [maxiter]: int, maximum number of iterations
            [lim]: float, smallest timescale to take into account [ms], if not None, the algorithm
                fits the slowest timescale first, then the next slowest, etc. !!Use only for
                decaying transfer functions!!
            [realpoles]: boolean, use real starting poles if true, use complex conjugate poles if
                false
            [initpoles]: 'lin' for linearly spaced initial poles, 'log10' and 'log' for
                logarithmically spaced poles
            [zerostart]: boolean, constrain the function to be 0 at t=0 if true
            [constrained]: fix the poles to be complex conjugate pairs
            [reduce_numexp]: boolean, pool short time scale exponentials together if true
            [return_real]: boolean, fix the output to only contain real poles if true
            [lower_limit]: float, if not None, lower limit on the exponential prefactors

        output:
            [alpha]: numpy array of (complex) timescales of exponentials
            [gamma]: numpy array of complex magnitudes of exponentials
            [pairs]: boolean array that indicates True at every index where a complex
                conjugate pair occurs
            [rms]: float, root mean square error
        """
        trialpoles, pairs = self._find_start_nodes(s, deg, realpoles, initpoles)

        if lim != None:
            a_s = []; c_s = []; pair_s = []
            y_decr = copy.copy(y)
            deg_decr = deg
            keep_going = True
            count = 0
            while keep_going:
                alist, clist, rmslist, pairslist = self._run_fit(s, y_decr, trialpoles, pairs, rtol, maxiter, constrained, zerostart)
                indmin = np.argmin(np.array(rmslist))
                anew, cnew, rmsnew, approx, pairsnew = self.reduceSeries(s, y_decr, alist[indmin], clist[indmin], pairs=pairslist[indmin], rtol=rtol)
                if count == 0:
                    # save parameters for later purposes
                    asave = copy.copy(anew)
                    csave = copy.copy(cnew)
                    rmssave = rmsnew
                    pairssave = pairsnew
                    surface_original = np.sum(cnew / anew)
                ind = []
                # take the longest timescale out
                ind.append(np.argmin(anew.real))
                if pairsnew[ind]:
                    ind.append(ind[0]+1)
                a_tba = anew[ind]
                c_tba = cnew[ind]
                pair_tba = pairsnew[ind]
                surface = np.sum(cnew / anew)
                y_old = copy.copy(y_decr)
                y_decr = self.sumFExp(s, anew, cnew) - self.sumFExp(s, anew[ind], cnew[ind])
                #~ deg_decr -= len(ind)
                trialpoles, pairs = self._find_start_nodes(s, deg_decr, True, initpoles)
                # stop if timescale is small enough
                if anew[ind][0] > 1e3 / (0.2*lim):
                    if len(ind) == 1:
                        c_tba = surface * a_tba
                    elif len(ind) == 2:
                        c_tba[0] = surface * (a_tba[0].real**2 + a_tba[0].imag**2) / (2.*a_tba[0].real)
                        c_tba[1] = c_tba[0]
                    else:
                        raise ValueError('invalid array length')
                    keep_going = False
                # stop if rmse is small enough
                approx = self.sumFExp(s, a_tba, c_tba)
                rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
                if rms < rtol:
                    keep_going = False
                # append new parameters to lists
                a_s += a_tba.tolist(); c_s += c_tba.tolist(); pair_s += pair_tba.tolist()
                # stop if to many parameters
                if count >= 9.:
                    keep_going = False
                count +=1

            # for returning
            alpha = np.array(a_s, dtype=complex); gamma = np.array(c_s, dtype=complex); pairs = np.array(pair_s, dtype=bool)
            approx = self.sumFExp(s, alpha, gamma)
            rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
            # check whether it was better to go with the first parameters
            if len(asave) < len(alpha) and rmssave < rtol:
                alpha = asave; gamma = csave; pairs = pairssave
                approx = self.sumFExp(s, alpha, gamma)
                rms = np.sqrt(((np.abs(approx-y) / np.max(np.abs(y)))**2).sum() / len(y))
            surface_after = np.sum(gamma / alpha)
            if np.abs(surface_original - surface_after) > rtol * surface_original:
                print('surface original: ', surface_original)
                print('surface after: ', surface_after)
            if np.min(alpha.real) < 0.:
                print('!!!owowow!!!')
        else:
            alist, clist, rmslist, pairslist = self._run_fit(s, y, trialpoles, pairs, rtol, maxiter, constrained, zerostart)
            indmin = np.argmin(np.array(rmslist))
            alpha, gamma, rms, approx, pairs = self.reduceSeries(s, y, alist[indmin], clist[indmin], pairs=pairslist[indmin], rtol=rtol)
            if reduce_numexp:
                alpha, gamma, rms, approx, pairs = self.reduceNumExp(s, y ,alpha, gamma, pairs, pplot=False)

        if return_real:
            inds = np.where(np.roll(pairs, 1))[0]
            alpha_real = np.delete(alpha.real, inds)
            alpha = alpha_real + 0j
            pairs = np.zeros(alpha.size, dtype=bool)
            gamma = self.fit_residues(s, y, alpha, pairs)

        if lower_limit:
            inds = np.where(alpha.real < lower_limit)[0]
            alpha = np.delete(alpha, inds)
            pairs = np.delete(pairs, inds)
            gamma = self.fit_residues(s, y, alpha, pairs)

        return alpha, gamma, pairs, rms

    def fitFExp_vector(self, s, ys, deg=20, rtol=1e-2, maxiter=5, extra_startpoles=[], extra_startpoles_pairs=[],
                        realpoles=True, initpoles='lin', reduce_series=False):
        """
        Fit multiple data-arrays in Fourrier-domain simultaneously with a shared set of nodes

        input:
            [s]: numpy array of complex number, frequencies of data
            [ys]: numpy ndarray of complex numbers, rows are different data-arrays
            [deg]: int, the starting number of nodes
            [rtol]: float, the relative tolercance at which to stop
            [maxiter]: int, the maximal number of iterations after which to stop when rtol
                is not reached
            [extra_startpoles]: numpy array of complex number, additional initial poles
            [extra_startpoles_pairs]: numpy bolean array, indicates complex conjugate pairs
                associated with the extra initial poles
            [realpoles]: boolean, if True the starting poles are real, if false the starting
                poles are complex conjugates (and then the real degree is 2*deg)
            [initpoles]: string specifying how the initial poles are distributed, choices are
                'lin', 'log' and 'log10'
            [reduce_series]: boolean, whether to delete expontentials of small influence after
                the fitting

        output:
            [alpha]: complex numpy array of exponential coefficients
            [gamma]: 2d complex numpy array, each row contains the residues corresponding to
                the respective data arrays
            [pairs]: boolean numpy array, indicates where a pair of complex conjugate
                exponentials occurs
            [rms]: float, aggregated root mean square error
        """
        trialpoles, pairs = self._find_start_nodes(s, deg, realpoles, initpoles)
        if len(extra_startpoles) > 0:
            trialpoles = np.concatenate((trialpoles, extra_startpoles))
            pairs = np.concatenate((pairs, extra_startpoles_pairs))
        alist, clist, rmslist, pairslist = self._run_fit_vector(s, ys, trialpoles, pairs, rtol, maxiter)
        indmin = np.argmin(np.array(rmslist))
        if reduce_series:
            # reduce the number of exponentials for each function separately
            alpha_arr = np.array([], dtype=complex)
            rms = 0.
            for ind, c in enumerate(clist[indmin]):
                alpha, gamma, rms_ind, approx, pair = self.reduceSeries(s, ys[ind], alist[indmin], c, pairs=pairslist[indmin], rtol=rtol)
                rms += rms_ind
                alpha_arr = np.concatenate((alpha_arr, alpha))
            alpha_arr = np.unique(alpha_arr)
            # search positions of common alphas
            asortind = np.argsort(alist[indmin])
            alphapos = np.searchsorted(alist[indmin][asortind], alpha_arr)
            inds = asortind[alphapos]
            return alist[indmin][inds], clist[indmin][:,inds], pairslist[indmin][inds], rms
        else:
            return alist[indmin], clist[indmin], pairslist[indmin], rmslist[indmin]


def create_logspace_freqarray(fmax=7, base=10, num=200):
    a = np.logspace(1, fmax, num=num, base=base)
    b = np.linspace(-base, base, num=num//2 + 1)
    # b = np.linspace(-base, base, num=num/2+(num/2)%2)[:-1]
    return 1j * np.concatenate((-a[::-1], b[1:-1], a))


class FourrierTools(object):
    """
    Performs an accurate Fourrier transform on functions
    evaluated at a given array of temporal grid points

    Parameters
    ----------
    tarr: `np.array` of floats,
        the time points (ms) at which the function is evaluated, have to be
        regularly spaced
    fmax: float, optional (default ``7.``)
        the maximum value to which the logarithm is evaluated to get the
        maximum evaluation frequency
    base: float, optional (defaul ``10``)
        the base of the logarithm used to generated the logspace
    num: int, even, optional (default ``200``)
        Number of points. the eventual number of points in frequency space
        is (2+1/2)*num

    Attributes
    ----------
    s: np.array of complex
        The frequencies at which input arrays in the Fourrier domain are
        supposed to be evaluated
    t: np.array of real
        The time array at which input arrays in the time domain are supposed
        to be evaluated
    ind_0s: int
        Index of the zero frequency component in `self.s`
    """
    def __init__(self, tarr, fmax=7, base=10, num=200):
        assert num % 2 == 0
        # create the frequency points at which to evaluate the transform
        self.s = create_logspace_freqarray(fmax=fmax, base=base, num=num)
        self.t = tarr
        self.ind_0s = len(self.s) // 2
        # create the quadrature matrix
        self._setQuad()
        self._setQuadInv()

    def _setQuad(self):
        s = self.s
        t = self.t
        N = len(t)
        dt = (t[1] - t[0])*1e-3
        c = np.zeros((len(s), N), dtype=complex)
        Nr = np.arange(1,N-1)[np.newaxis,:]
        sc = s[:,np.newaxis]
        mask_arr = np.abs(sc)>1e-12
        # first frequency integral
        np.divide(np.exp(-sc*dt) - 1., -sc*dt, out=c[:,0:1], where=mask_arr)
        np.divide(-1. + c[:,0:1], -sc, out=c[:,0:1], where=mask_arr)
        c[np.where(np.logical_not(mask_arr))[0],0] = dt / 2.
        # middle integrals
        np.divide(np.exp(-sc*dt*(Nr+1)) - 2.*np.exp(-sc*dt*Nr) + np.exp(-sc*dt*(Nr-1)), sc**2 * dt,
                                out=c[:,1:-1], where=mask_arr)
        c[np.where(np.logical_not(mask_arr))[0],1:-1] = dt
        # last frequency integral
        np.divide(np.exp(-sc*dt*N) - np.exp(-sc*dt*(N-1)), -sc*dt, out=c[:,N-1:N], where=mask_arr)
        np.divide(np.exp(-sc*dt*N) - c[:,N-1:N], -sc, out=c[:,N-1:N], where=mask_arr)
        c[np.where(np.logical_not(mask_arr))[0],N-1] = dt / 2.
        self.c = c

    def _setQuadInv(self):
        t = self.t[:,np.newaxis]*1e-3
        s = self.s[np.newaxis,:]
        ic = np.zeros((len(self.t), len(self.s)), dtype=complex)
        mask_arr = np.abs(t)>1e-12
        # compute integrals
        I1 =  np.divide(np.exp(s[:,1:]*t) - np.exp(s[:,:-1]*t), 1j*t,
                        out=np.zeros_like(ic[:,:-1]), where=mask_arr)
        I1[np.where(np.logical_not(mask_arr))[0],:] = (s[:,1:] - s[:,:-1]) / 1j
        I2_ = np.divide(np.exp(s[:,1:]*t) - np.exp(s[:,:-1]*t), 1j*t,
                        out=np.zeros_like(ic[:,:-1]), where=mask_arr)
        # I2_[0,:] = s[:,1:] - s[:,:-1] / 1j
        I2 =  np.divide((s[:,1:]-s[:,:-1]).imag * np.exp(s[:,1:]*t) - I2_, 1j*t,
                        out=np.zeros_like(ic[:,:-1]), where=mask_arr)
        I2[0:1,:] = s[:,1:] * (s[:,1:] - s[:,:-1])
        # compute matrix elements
        ic[:,0] = I1[:,0] - I2[:,0] / (s[:,1] - s[:,0]).imag
        ic[:,1:-1] = I1[:,1:] - I2[:,1:] / (s[:,2:] - s[:,1:-1]).imag + I2[:,:-1] / (s[:,1:-1] - s[:,:-2]).imag
        ic[:,-1] = I2[:,-1] / (s[:,-1] - s[:,-2]).imag
        self.ic = ic / (2.*np.pi)

    def __call__(self, arr):
        """
        Evaluate the Fourrier transform of `arr`

        Parameters
        ----------
            arr: `np.array`
                Should have the same length as `self.t`

        Returns
        -------
            s: `np.array`
                the frequency points at which the Fourrier
                transform is evaluated (in Hz)
            farr: `np.array`
                the Fourrier transform of `arr`
        """
        farr = np.dot(self.c, arr)
        return self.s, farr

    def ft(self, arr):
        """
        Evaluate the Fourrier transform of `arr`

        Parameters
        ----------
            arr: `np.array`
                Should have the same length as `self.t`

        Returns
        -------
            s: `np.array`
                the frequency points at which the Fourrier
                transform is evaluated (in Hz)
            farr: `np.array`
                the Fourrier transform of `arr`
        """
        return self(arr)

    def ftInv(self, arr):
        """
        Evaluate the inverse Fourrier transform of `arr`

        Parameters
        ----------
            arr: `np.array`
                Should have the same length as `self.s`

        Returns
        -------
            t: `np.array`
                the time points at which the inverse Fourrier
                transform is evaluated (in ms)
            tarr: `np.array`
                the Fourrier transform of `arr`
        """
        tarr = np.dot(self.ic, arr)
        return self.t, tarr


class expExtractor(object):
    def __call__(self, N=1, recalc=False, atol=5e-2, pprint=True):
        """
        Fit a partial fraction decomposition to the obtained kernel in the frequency domain

        input:
            N (int): the number of exponenstials to use
            recalc (bool): whether to force a refit if a fit with the given N already exists
            atol (float): the tolerance, if the RMSE of the fit is higher
                and pprint is True a warning is printed
            pprint (bool): whether to print the warning

        output:
            kfit (dict): {'a': array of the exponental factors (1/tau), 'c': array of the
                coefficients, 'p': idicators whether or not a given exponential is part
                of a complex conjugate pair}
        """
        if N not in self.kfit or recalc:
            FEF = fExpFitter()
            ak, ck, pk, rms = FEF.fitFExp(self.s_f, self.k_f, rtol=1e-2, deg=N, maxiter=20,
                            initpoles='log10', realpoles=True, zerostart=False, constrained=True, reduce_numexp=False)
            ak *= 1e-3; #ck *= 1e-3  # convert units to ms
            # also try time domain approach
            if N < 10:
                EF = ExpFitter()
                if 'k_t' not in self.__dict__:
                    k_t = EF.sumExp(self.tarr, -self.kfit[30]['a'], self.kfit[30]['c'])
                else:
                    k_t = self.k_t
                ak_, ck_, rms_ = EF.PronyExpFit(N, self.tarr, k_t)
                ak_ = -ak_
                pk_ = np.zeros(ak_.shape, dtype=bool)
                # go for the approach with the lowest error
                rms = np.sqrt(((k_t - EF.sumExp(self.tarr, -ak, ck))**2).sum() / len(k_t))
                if rms > rms_:
                    ak = ak_; ck = ck_; pk = pk_; rms = rms_
            if rms > atol and pprint:
                print('No sane fit achieved for N=' + str(N))
                print('RMSE:', rms)
            self.kfit[N] = {'a': ak, 'c': ck, 'p': pk}
        return self.kfit[N]

    def fit_vector(self, N, atol=5e-2, pprint=True, store=True):
        FEF = fExpFitter()
        ak, ck, pk, rms = FEF.fitFExp(self.s_f, self.k_f, rtol=1e-2, deg=N, maxiter=20,
                        initpoles='log10', realpoles=True, zerostart=False, constrained=True, reduce_numexp=False)
        ak *= 1e-3; #ck *= 1e-3  # convert units to ms
        if rms > atol and pprint:
            print('No sane fit achieved for N=' + str(N))
            print('RMSE:', rms)
        res = {'a': ak, 'c': ck, 'p': pk}
        if store:
            self.kfit[N] = res
        return res

    def fit_prony(self, N, atol=5e-2, pprint=True, store=True):
        EF = ExpFitter()
        if 'k_t' not in self.__dict__:
            k_t = EF.sumExp(self.tarr, -self.kfit[30]['a'], self.kfit[30]['c'])
        else:
            k_t = self.k_t
        ak, ck, rms = EF.PronyExpFit(N, self.tarr, k_t)
        ak *= -1.
        # ck *= 1e-3
        pk = np.zeros(ak.shape, dtype=bool)
        if rms > atol and pprint:
            print('No sane fit achieved for N=' + str(N))
            print('RMSE:', rms)
        res = {'a': ak, 'c': ck, 'p': pk}
        if store:
            self.kfit[N] = res
        return res

    def k_freq(self, N):
        self(N)
        FEF = fExpFitter()
        return FEF.sumFExp(self.s_f, self.kfit[N]['a']*1e3, self.kfit[N]['c'])

    def k_time(self, N):
        self(N)
        EF = ExpFitter()
        return EF.sumExp(self.tarr, -self.kfit[N]['a'], self.kfit[N]['c'])


class IzExtractor(expExtractor):
    def __init__(self, Iz, yarr):
        self.tarr = Iz
        FT = FourrierTools(self.tarr)
        self.s_f, self.k_f = FT(yarr)
        # for storing kernel fit results
        self.kfit = {}
        # initial fit
        self(30)


class simpleExpExtractor(expExtractor):
    def __init__(self, arr, tarr):
        dt = tarr[1] - tarr[0]
        self.tarr = tarr
        self.k_t = arr
        # create a smoothing window if arr hasn't gone to zero yet
        if arr[-1] > 1e-9:
            vwindow = np.cos((np.pi/2.)*(self.tarr/self.tarr[-1]))
        else:
            vwindow = np.ones(self.tarr.shape)
        # compute Fourrier transform
        FT = FourrierTools(self.tarr, fmax=np.log10(1./(dt*1e-3)), base=10., num=200)
        self.s_f, self.k_f = FT(self.k_t)
        # for storing kernel fit results
        self.kfit = {}
        # initial fit
        self(30)


class kernelExtractor(expExtractor):
    """
    This class computes the rescale kernel between the
    synapse in the dendrite and the synapse in the soma.

    input:
        vdend (numpy 1d array): the somatic voltage recording
            with the synapse in the dendrites
        vsoma (numpy 1d array): the somatic voltage recording
            with the synapse at the some
        vbase (numpy 1d array): the somatic voltage recording
            without the synapse
        time  (numpy 1d array): the time array associated with
            the recordings

    attributes:
        vdend: somatic voltage from dendritic spike arrival onwards
        vsoma: somatic voltage from somatic spike arrival onwards
        vbase: somatic voltage without spike arrival
        vdend_: vdend - vbase
        vsoma_: vsoma - vbase
        s_f: frequency array
        vdend_f: fourier transform of vdend_
        vdend_f: fourier transform of vdend_
        k_f: kernel in the frequency domain
    """
    def __init__(self, vdend, vsoma, vbase, time):
        # time step
        dt = time[1] - time[0]
        # voltage deviations from baseline
        vdend_ = vdend-vbase
        vsoma_ = vsoma-vbase
        # find when the spike arrives
        istart = np.where(vdend_ > 1e-9)[0][0] - 1.
        # time array PSP
        self.tarr = time[istart:] - istart*dt
        # full potentials after spike arrival
        self.vdend = vdend[istart:]
        self.vsoma = vsoma[istart:]
        self.vbase = vbase[istart:]
        # create a smoothing window if PSP hasn't gone to zero yet
        if vdend_[-1] > 1e-9:
            vwindow = np.cos((np.pi/2.)*(self.tarr/self.tarr[-1]))
        else:
            vwindow = np.ones(self.tarr.shape)
        # PSPs
        self.vdend_ = vdend_[istart:] * vwindow
        self.vsoma_ = vsoma_[istart:] * vwindow
        # fourrier transform
        FT = FourrierTools(self.tarr, fmax=np.log10(1./(dt*1e-3)), base=10., num=200)
        self.s_f, self.vdend_f = FT(self.vdend_)
        self.s_f, self.vsoma_f = FT(self.vsoma_)
        # compute the kernel
        self.k_f = self.vdend_f / self.vsoma_f
        # for storing kernel fit results
        self.kfit = {}
        # initial fit
        self(30)

class expNReducer(expExtractor):
    def __init__(self, alphas, gammas):
        FEF = fExpFitter()
        # creat time array
        self.tarr = np.linspace(0.,100.,1e3)
        # create frequency array
        self.s_f = create_logspace_freqarray(fmax=7, base=10, num=200)
        # kernel in freqency domain
        self.k_f = FEF.sumFExp(self.s_f, -alphas*1e3, gammas*1e3)
        # for storing kernel fit results
        self.kfit = {}
        # initial fit
        self(30)





"""
File contains:

    - `neat.GreensNode`
    - `neat.SomaGreensNode`
    - `neat.GreensTree`

Author: W. Wybo
"""

import numpy as np

import copy

from . import morphtree
from .morphtree import MorphLoc
from .phystree import PhysNode, PhysTree


class GreensNode(PhysNode):
    '''
    Node that stores quantities and defines functions to implement the impedance
    matrix calculation based on Koch's algorithm (Koch & Poggio, 1985).

    Attributes
    ----------
    expansion_points: dict {str: np.ndarray}
        Stores ion channel expansion points for this segment.
    '''
    def __init__(self, index, p3d):
        super().__init__(index, p3d)
        self.expansion_points = {}

    def _rescaleLengthRadius(self):
        self.R_ = self.R * 1e-4 # convert to cm
        self.L_ = self.L * 1e-4 # convert to cm

    def setExpansionPoint(self, channel_name, statevar):
        """
        Set the choice for the state variables of the ion channel around which
        to linearize.

        Note that when adding an ion channel to the node, the default expansion
        point setting is to linearize around the asymptotic values for the state
        variables at the equilibrium potential store in `self.e_eq`.
        Hence, this function only needs to be called to change that setting.

        Parameters
        ----------
        channel_name: string
            the name of the ion channel
        statevar: dict
            The expansion points for each of the ion channel state variables
        """
        if statevar is None:
            statevar = {}
        self.expansion_points[channel_name] = statevar

    def getExpansionPoint(self, channel_name):
        try:
            return self.expansion_points[channel_name]
        except KeyError:
            self.expansion_points[channel_name] = {}
            return self.expansion_points[channel_name]

    def _calcMembraneImpedance(self, freqs, channel_storage, use_conc=False):
        """
        Compute the impedance of the membrane at the node

        Parameters
        ----------
        freqs: `np.ndarray` (``dtype=complex``, ``ndim=1``)
            The frequencies at which the impedance is to be evaluated
        channel_storage: dict of ion channels (optional)
            The ion channels that have been initialized already. If not
            provided, a new channel is initialized
        use_conc: bool
            if True, also uses concentration mechanisms to compute linearized
            membrane impedance

        Returns
        -------
        `np.ndarray` (``dtype=complex``, ``ndim=1``)
            The membrane impedance
        """
        if use_conc:
            g_m_ions = {conc: np.zeros_like(freqs) for conc in list(self.concmechs.keys())}
        g_m_aux = self.c_m * freqs + self.currents['L'][0]
        # loop over channels that do not read concentrations
        for channel_name in set(self.currents.keys()) - set('L'):
            g, e = self.currents[channel_name]
            if g > 1e-10:
                # create the ionchannel object
                channel = channel_storage[channel_name]
                if len(channel.conc) == 0:
                    # check if needs to be computed around expansion point
                    sv = self.getExpansionPoint(channel_name)
                    # add channel contribution to membrane impedance
                    g_ = g * channel.computeLinSum(self.e_eq, freqs, e, **sv)
                    g_m_aux -= g_
                    if use_conc:
                        g_m_ions[channel.ion] += g_
                        # g_m_ions[channel.ion] += g * channel.computePOpen(self.e_eq, statevars=sv)
        # loop over channels that do read concentrations
        for channel_name in set(self.currents.keys()) - set('L'):
            g, e = self.currents[channel_name]
            if g > 1e-10:
                # create the ionchannel object
                channel = channel_storage[channel_name]
                if len(channel.conc) > 0:
                    # check if needs to be computed around expansion point
                    sv = self.getExpansionPoint(channel_name)
                    # add channel contribution to membrane impedance
                    g_m_aux -= g * channel.computeLinSum(self.e_eq, freqs, e, **sv)
                    if use_conc:
                        for ion in channel.conc:
                            g_m_aux -= g * channel.computeLinConc(self.e_eq, freqs, e, ion) * \
                                       self.concmechs[ion].computeLinear(freqs) * \
                                       g_m_ions[ion]

        return 1. / (2. * np.pi * self.R_ * g_m_aux)

    def _setImpedance(self, freqs, channel_storage, use_conc=False):
        self.counter = 0
        self.z_m = self._calcMembraneImpedance(freqs, channel_storage,
                                              use_conc=use_conc)
        self.z_a = self.r_a / (np.pi * self.R_**2)
        self.gamma = np.sqrt(self.z_a / self.z_m)
        self.z_c = self.z_a / self.gamma

    def _setImpedanceDistal(self):
        """
        Set the boundary condition at the distal end of the segment
        """
        if len(self.child_nodes) == 0:
            self.z_distal = np.infty*np.ones(len(self.z_m)) if self.g_shunt < 1e-10 else \
                            1. / self.g_shunt
        else:
            self.z_distal = 1. / (np.sum([1. / cnode._collapseBranchToRoot() \
                                         for cnode in self.child_nodes], 0) + \
                                  self.g_shunt)

    def _setImpedanceProximal(self):
        """
        Set the boundary condition at the proximal end of the segment
        """
        # child nodes of parent node without the current node
        sister_nodes = copy.copy(self.parent_node.child_nodes[:])
        sister_nodes.remove(self)
        # compute the impedance
        val = 0.
        if self.parent_node is not None:
            val += 1. / self.parent_node._collapseBranchToLeaf()
            val += self.parent_node.g_shunt
        for snode in sister_nodes:
            val += 1. / snode._collapseBranchToRoot()
        self.z_proximal = 1. / val

    def _collapseBranchToLeaf(self):
        return self.z_c * (self.z_proximal * np.cosh(self.gamma * self.L_) + \
                           self.z_c * np.sinh(self.gamma * self.L_)) / \
                          (self.z_proximal * np.sinh(self.gamma * self.L_) +
                           self.z_c * np.cosh(self.gamma * self.L_))

    def _collapseBranchToRoot(self):
        zr = self.z_c * (np.cosh(self.gamma * self.L_) +
                         self.z_c / self.z_distal * np.sinh(self.gamma * self.L_)) / \
                        (np.sinh(self.gamma * self.L_) +
                         self.z_c / self.z_distal * np.cosh(self.gamma * self.L_))
        return zr

    def _setImpedanceArrays(self):
        self.gammaL = self.gamma * self.L_
        self.z_cp = self.z_c / self.z_proximal
        self.z_cd = self.z_c / self.z_distal
        self.wrongskian = np.cosh(self.gammaL) / self.z_c * \
                           (self.z_cp + self.z_cd + \
                            (1. + self.z_cp * self.z_cd) * np.tanh(self.gammaL))
        self.z_00 = (self.z_cd * np.sinh(self.gammaL) + np.cosh(self.gammaL)) / \
                    self.wrongskian
        self.z_11 = (self.z_cp * np.sinh(self.gammaL) + np.cosh(self.gammaL)) / \
                    self.wrongskian
        self.z_01 = 1. / self.wrongskian

    def _calcZF(self, x1, x2):
        if x1 <  1e-3 and x2 < 1e-3:
            return self.z_00
        elif x1 > 1.-1e-3 and x2 > 1.-1e-3:
            return self.z_11
        elif (x1 < 1e-3 and x2 > 1.-1e-3) or (x1 > 1.-1e-3 and x2 < 1e-3):
            return self.z_01
        elif x1 < x2:
            return (self.z_cp * np.sinh(self.gammaL*x1) + np.cosh(self.gammaL*x1)) * \
                   (self.z_cd * np.sinh(self.gammaL*(1.-x2)) + np.cosh(self.gammaL*(1.-x2))) / \
                   self.wrongskian
        else:
            return (self.z_cp * np.sinh(self.gammaL*x2) + np.cosh(self.gammaL*x2)) * \
                   (self.z_cd * np.sinh(self.gammaL*(1.-x1)) + np.cosh(self.gammaL*(1.-x1))) / \
                   self.wrongskian


class SomaGreensNode(GreensNode):
    def _calcMembraneImpedance(self, freqs, channel_storage, use_conc=False):
        z_m = super()._calcMembraneImpedance(freqs, channel_storage,
                                                                use_conc=use_conc)
        # rescale for soma surface instead of cylinder radius
        # return z_m / (2. * self.R_)
        return 1. / (2. * self.R_ / z_m + self.g_shunt)

    def _setImpedance(self, freqs, channel_storage, use_conc=False):
        self.counter = 0
        self.z_soma = self._calcMembraneImpedance(freqs, channel_storage,
                                                 use_conc=use_conc)

    def _collapseBranchToLeaf(self):
        return self.z_soma

    def _setImpedanceArrays(self):
        val = 1. / self.z_soma
        for node in self.child_nodes:
            val += 1. / node._collapseBranchToRoot()
        self.z_in = 1. / val

    def _calcZF(self, x1, x2):
        return self.z_in


class GreensTree(PhysTree):
    """
    Class that computes the Green's function in the Fourrier domain of a given
    neuronal morphology (Koch, 1985). This three defines a special
    `neat.SomaGreensNode` as a derived class from `neat.GreensNode` as some
    functions required for Green's function calculation are different and thus
    overwritten.

    The calculation proceeds on the computational tree (see docstring of
    `neat.MorphNode`). Thus it makes no sense to look for Green's function
    related quantities in the original tree.

    Attributes
    ----------
    freqs: np.array of complex
        Frequencies at which impedances are evaluated
    """
    def __init__(self, file_n=None, types=[1,3,4]):
        super().__init__(file_n=file_n, types=types)
        self.freqs = None

    def _createCorrespondingNode(self, node_index, p3d=None):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
        node_index: `int`
            index of the new node
        """
        if node_index == 1:
            return SomaGreensNode(node_index, p3d)
        else:
            return GreensNode(node_index, p3d)

    def removeExpansionPoints(self):
        """
        Remove expansion points from all nodes in the tree
        """
        for node in self:
            node.expansion_points = {}

    @morphtree.computationalTreetypeDecorator
    def setImpedance(self, freqs, use_conc=False, pprint=False):
        """
        Set the boundary impedances for each node in the tree

        Parameters
        ----------
        freqs: `np.ndarray` (``dtype=complex``, ``ndim=1``)
            frequencies at which the impedances will be evaluated [Hz]
        use_conc: bool
            whether or not to incorporate concentrations in the calculation
        pprint: bool (default ``False``)
            whether or not to print info on the progression of the algorithm

        """
        self.freqs = freqs
        # set the node specific impedances
        for node in self:
            node._rescaleLengthRadius()
            node._setImpedance(freqs, self.channel_storage, use_conc=use_conc)
        # recursion
        if len(self) > 1:
            self._impedanceFromLeaf(self.leafs[0], self.leafs[1:], pprint=pprint)
            self._impedanceFromRoot(self.root)
        # clean
        for node in self:
            node.counter = 0
            node._setImpedanceArrays()

    def _impedanceFromLeaf(self, node, leafs, pprint=False):
        if pprint:
            print('Forward sweep: ' + str(node))
        pnode = node.parent_node
        # log how many times recursion has passed at node
        if not self.isLeaf(node):
            node.counter += 1
        # if the number of childnodes of node is equal to the amount of times
        # the recursion has passed node, the distal impedance can be set. Otherwise
        # we start a new recursion at another leaf.
        if node.counter == len(node.child_nodes):
            node._setImpedanceDistal()
            if not self.isRoot(node):
                self._impedanceFromLeaf(pnode, leafs, pprint=pprint)
        elif len(leafs) > 0:
                self._impedanceFromLeaf(leafs[0], leafs[1:], pprint=pprint)

    def _impedanceFromRoot(self, node):
        if node != self.root:
            node._setImpedanceProximal()
        for cnode in node.child_nodes:
            self._impedanceFromRoot(cnode)

    @morphtree.computationalTreetypeDecorator
    def calcZF(self, loc1, loc2):
        """
        Computes the transfer impedance between two locations for all frequencies
        in `self.freqs`.

        Parameters
        ----------
        loc1: dict, tuple or `:class:MorphLoc`
            One of two locations between which the transfer impedance is computed
        loc2: dict, tuple or `:class:MorphLoc`
            One of two locations between which the transfer impedance is computed

        Returns
        -------
        nd.ndarray (dtype = complex, ndim = 1)
            The transfer impedance as a function of frequency
        """
        # cast to morphlocs
        loc1 = MorphLoc(loc1, self)
        loc2 = MorphLoc(loc2, self)
        # the path between the nodes
        path = self.pathBetweenNodes(self[loc1['node']], self[loc2['node']])
        # compute the kernel
        z_f = np.ones_like(self.freqs)
        if len(path) == 1:
            # both locations are on same node
            z_f *= path[0]._calcZF(loc1['x'], loc2['x'])
        else:
            # different cases whether path goes to or from root
            if path[1] == self[loc1['node']].parent_node:
                z_f *= path[0]._calcZF(loc1['x'], 0.)
            else:
                z_f *= path[0]._calcZF(loc1['x'], 1.)
                z_f /= path[0]._calcZF(1., 1.)
            if path[-2] == self[loc2['node']].parent_node:
                z_f *= path[-1]._calcZF(loc2['x'], 0.)
            else:
                z_f *= path[-1]._calcZF(loc2['x'], 1.)
                z_f /= path[-1]._calcZF(1., 1.)
            # nodes within the path
            ll = 1
            for node in path[1:-1]:
                z_f /= node._calcZF(1., 1.)
                if path[ll-1] not in node.child_nodes or \
                   path[ll+1] not in node.child_nodes:
                    z_f *= node._calcZF(0., 1.)
                ll += 1

        return z_f

    @morphtree.computationalTreetypeDecorator
    def calcImpedanceMatrix(self, locarg, explicit_method=True):
        """
        Computes the impedance matrix of a given set of locations for each
        frequency stored in `self.freqs`.

        Parameters
        ----------
        locarg: `list` of locations or string
            if `list` of locations, specifies the locations for which the
            impedance matrix is evaluated, if ``string``, specifies the
            name under which a set of location is stored
        explicit_method: bool, optional (default ``True``)
            if ``False``, will use the transitivity property of the impedance
            matrix to further optimize the computation.

        Returns
        -------
        `np.ndarray` (``dtype = self.freqs.dtype``, ``ndim = 3``)
            the impedance matrix, first dimension corresponds to the
            frequency, second and third dimensions contain the impedance
            matrix at that frequency
        """
        if isinstance(locarg, list):
            locs = [MorphLoc(loc, self) for loc in locarg]
        elif isinstance(locarg, str):
            locs = self.getLocs(locarg)
        else:
            raise IOError('`locarg` should be list of locs or string')
        z_mat = np.zeros((len(self.freqs), len(locs), len(locs)),
                         dtype=self.freqs.dtype)

        if explicit_method:
            for ii, loc0 in enumerate(locs):
                jj = 0
                while jj < ii:
                    loc1 = locs[jj]
                    z_f = self.calcZF(loc0, loc1)
                    z_mat[:,ii,jj] = z_f
                    z_mat[:,jj,ii] = z_f
                    jj += 1
                z_f = self.calcZF(loc0, loc0)
                z_mat[:,ii,ii] = z_f

        else:
            for ii in range(len(locs)):
                self._calcImpedanceMatrixFromNode(ii, locs, z_mat)

        return z_mat

    def _calcImpedanceMatrixFromNode(self, ii, locs, z_mat):
        node = self[locs[ii]['node']]
        for jj, loc in enumerate(locs):
            if loc['node'] == node.index and jj >= ii:
                z_new = node._calcZF(locs[ii]['x'],loc['x'])
                z_mat[:,ii,jj] = z_new
                z_mat[:,jj,ii] = z_new
        # move down
        for c_node in node.child_nodes:
            z_new = node._calcZF(locs[ii]['x'], 1.)
            self._calcImpedanceMatrixDown(ii, z_new, c_node, locs, z_mat)

        if node.parent_node is not None:
            z_new = node._calcZF(locs[ii]['x'], 0.)
            # move to sister nodes
            for c_node in set(node.parent_node.child_nodes) - {node}:
                self._calcImpedanceMatrixDown(ii, z_new, c_node, locs, z_mat)
            # move up
            self._calcImpedanceMatrixUp(ii, z_new, node.parent_node, locs, z_mat)

    def _calcImpedanceMatrixUp(self, ii, z_0, node, locs, z_mat):
        # compute impedances
        z_in = node._calcZF(1.,1.)
        for jj, loc in enumerate(locs):
            if jj > ii and loc['node'] == node.index:
                z_new = z_0 / z_in * node._calcZF(1.,loc['x'])
                z_mat[:,ii,jj] = z_new
                z_mat[:,jj,ii] = z_new

        if node.parent_node is not None:
            z_new = z_0 / z_in * node._calcZF(0., 1.)
            # move to sister nodes
            for c_node in set(node.parent_node.child_nodes) - {node}:
                self._calcImpedanceMatrixDown(ii, z_new, c_node, locs, z_mat)
            # move to parent node
            z_new = z_0 / z_in * node._calcZF(0., 1.)
            self._calcImpedanceMatrixUp(ii, z_new, node.parent_node, locs, z_mat)

    def _calcImpedanceMatrixDown(self, ii, z_0, node, locs, z_mat):
        # compute impedances
        z_in = node._calcZF(0.,0.)
        for jj, loc in enumerate(locs):
            if jj > ii and loc['node'] == node.index:
                z_new = z_0 / z_in * node._calcZF(0., loc['x'])
                z_mat[:,ii,jj] = z_new
                z_mat[:,jj,ii] = z_new

        # recurse to child nodes
        z_new = z_0 / z_in * node._calcZF(0., 1.)
        for c_node in node.child_nodes:
            self._calcImpedanceMatrixDown(ii, z_new, c_node, locs, z_mat)




